/* vi: set sw=4 ts=4: */
/*
 *	xstream.c
 *
 *	extensible stream module.
 *
 *	Created by David Hsieh <david_hsieh@alphanetworks.com>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>

#include <dtrace.h>
#include <mem_helper.h>

#include "xstream.h"

#define MAX_XS_UNDO		512	/* the maximum number of the unget() function can be accept. */
#define XSTREAM_T(x)	(struct xstream *)(x)

struct xsbuff
{
	unsigned char * start;
	size_t curr, size;
};

struct xstream
{
	xstype_t			type;
	union
	{
		FILE *			file;
		int				sock;
		struct xsbuff	buff;
	}					fd;
	unsigned char		undo[MAX_XS_UNDO];
	size_t				udsize;
};

/* open file type stream */
xstream_t xs_fopen(const char * file, const char * mode)
{
	FILE * fd;
	struct xstream * xs;

	fd = fopen(file, mode);
	if (fd)
	{
		xs = xmalloc(sizeof(struct xstream));
		if (xs)
		{
			memset(xs, 0, sizeof(struct xstream));
			xs->type = XSTYPE_FILE;
			xs->fd.file = fd;
			return xs;
		}
		fclose(fd);
	}
	return NULL;
}

/* open socket type stream */
xstream_t xs_sopen(int domain, int type, int protocol)
{
	int sd;
	struct xstream * xs;

	sd = socket(domain, type, protocol);
	if (sd > 0)
	{
		xs = xmalloc(sizeof(struct xstream));
		if (xs)
		{
			memset(xs, 0, sizeof(struct xstream));
			xs->type = XSTYPE_SOCKET;
			xs->fd.sock = sd;
			return xs;
		}
		close(sd);
	}
	return NULL;
}

/* open buffer type stream */
xstream_t xs_bopen(void * buff, size_t size)
{
	struct xstream * xs;
	if (buff && size)
	{
		xs = xmalloc(sizeof(struct xstream));
		if (xs)
		{
			memset(xs, 0, sizeof(struct xstream));
			xs->type = XSTYPE_BUFFER;
			xs->fd.buff.start = (unsigned char *)buff;
			xs->fd.buff.curr = 0;
			xs->fd.buff.size = size;
			return xs;
		}
	}
	return NULL;
}

int xs_close(xstream_t fd)
{
	struct xstream * xs = XSTREAM_T(fd);
	if (!xs) return -1;
	switch (xs->type)
	{
	case XSTYPE_FILE:	fclose(xs->fd.file);	break;
	case XSTYPE_SOCKET:	close(xs->fd.sock);		break;
	case XSTYPE_BUFFER:	break;
	case XSTYPE_UNKNOWN: break;
	}
	xfree(xs);
	return 0;
}

int xs_getc(xstream_t fd)
{
	struct xstream * xs = XSTREAM_T(fd);
	int c = EOF;
	if (!xs) return c;
	if (xs->udsize > 0)
	{
		xs->udsize--;
		c = (int)(xs->undo[xs->udsize]);
		return c;
	}
	switch (xs->type)
	{
	case XSTYPE_FILE:	c = fgetc(xs->fd.file);	break;
	case XSTYPE_SOCKET:	/* TODO: */ break;
	case XSTYPE_BUFFER:
		if (xs->fd.buff.curr < xs->fd.buff.size)
		{
			c = (int)xs->fd.buff.start[xs->fd.buff.curr];
			xs->fd.buff.curr++;
		}
		break;

	default:
		c = EOF;
		break;
	}
	return c;
}

int xs_ungetc(int c, xstream_t fd)
{
	struct xstream * xs = XSTREAM_T(fd);

	if (!xs) return EOF;
	if (c==EOF) return EOF;
	if (xs->udsize < MAX_XS_UNDO)
	{
		xs->undo[xs->udsize] = (unsigned char)c;
		xs->udsize++;
		return c;
	}
	return EOF;
}

int xs_ungets(const char * s, xstream_t fd)
{
	int c;
	size_t len;

	if (!s) return 0;
	len = strlen(s);
	while (len>0)
	{
		c = s[--len];
		if (xs_ungetc(c, fd)==EOF) return EOF;
	}
	return 0;
}
