/* vi: set sw=4 ts=4: */
/*
 *	elbox_physmap.c
 *
 *	flash mapping for elbox.
 *	by David Hsieh <david_hsieh@alphanetworks.com>
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/proc_fs.h>
#include <asm/io.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/map.h>

#ifndef CONFIG_MTD_PARTITIONS
#error "We need CONFIG_MTD_PARTITIONS !!!"
#else
#include <linux/mtd/partitions.h>
#include <linux/squashfs_fs.h>
#endif

#define WINDOW_ADDR	CONFIG_PHYSMAP_ELBOX_START
#define WINDOW_SIZE	CONFIG_PHYSMAP_ELBOX_LEN
#define BUSWIDTH	CONFIG_PHYSMAP_ELBOX_BUSWIDTH
#define KERNEL_SKIP	CONFIG_ELBOX_KERNEL_SKIP

static struct mtd_info *mymtd = NULL;

static __u8 _map_read8(struct map_info *map, unsigned long ofs)
{
	return __raw_readb(map->map_priv_1 + ofs);
}

static __u16 _map_read16(struct map_info *map, unsigned long ofs)
{
	return __raw_readw(map->map_priv_1 + ofs);
}

static __u32 _map_read32(struct map_info *map, unsigned long ofs)
{
	return __raw_readl(map->map_priv_1 + ofs);
}

static void _map_copy_from(struct map_info *map, void *to, unsigned long from, ssize_t len)
{
	memcpy_fromio(to, map->map_priv_1 + from, len);
}

static void _map_write8(struct map_info *map, __u8 d, unsigned long adr)
{
	__raw_writeb(d, map->map_priv_1 + adr);
	mb();
}

static void _map_write16(struct map_info *map, __u16 d, unsigned long adr)
{
	__raw_writew(d, map->map_priv_1 + adr);
	mb();
}

static void _map_write32(struct map_info *map, __u32 d, unsigned long adr)
{
	__raw_writel(d, map->map_priv_1 + adr);
	mb();
}

static void _map_copy_to(struct map_info *map, unsigned long to, const void *from, ssize_t len)
{
	memcpy_toio(map->map_priv_1 + to, from, len);
}

static struct map_info physmap_map =
{
	name:		"ELBOX Physically mapped flash",
	size:		WINDOW_SIZE,
	buswidth:	BUSWIDTH,
	read8:		_map_read8,
	read16:		_map_read16,
	read32:		_map_read32,
	copy_from:	_map_copy_from,
	write8:		_map_write8,
	write16:	_map_write16,
	write32:	_map_write32,
	copy_to:	_map_copy_to
};



/*
 *	Flash layout
 *
 *	|      16K      |      16K      |      16K      |      16K      |
 *	+---------------------------------------------------------------+
 *	|                                                               |
 *	| Boot code (64K x 3)                                           |
 *	|                                                               |
 *	+---------------------------------------------------------------+
 *	| Boot Config & Board Data & RF parameters                      |
 *	+---------------------------------------------------------------+
 *	|                                                               |
 *	| Upgrade (linux kernel/rootfs)                                 |
 *	~                                                               ~
 *	|                                                               |
 *	|                                                               |
 *	+---------------------------------------------------------------+
 *	| Language pack (64K x 1)                                       |
 *	+---------------------------------------------------------------+
 *	| Board Config (64K x 1)                                        |
 *	+---------------------------------------------------------------+
 */
#define SIZE_BOOTCODE	0x30000
#ifdef CONFIG_MTD_LANG_PACK
#define SIZE_LANGPACK	CONFIG_MTD_LANG_PACK_SIZE
#else
#define SIZE_LANGPACK	0
#endif
#define SIZE_RGDB		0x10000	/* Size of Board Configuration (rgdb) */
#define SIZE_BDCFG		0x10000	/* Size of Board Data (bdcfg) */

/* upgradeL this block is used to store the runtime image. */
#define UPGRADE_OFFSET	SIZE_BOOTCODE + SIZE_BDCFG
#define UPGRADE_SIZE	(WINDOW_SIZE - SIZE_BOOTCODE - SIZE_LANGPACK - SIZE_RGDB - SIZE_BDCFG)

/* rgdb: this block is used to store the user configuration. */
#define RGDB_OFFSET		(WINDOW_SIZE - SIZE_RGDB)
#define RGDB_SIZE		SIZE_RGDB

/* bdcfg: this block is used to soitre the board configuration. */
#define BDCFG_OFFSET	SIZE_BOOTCODE
#define BDCFG_SIZE		SIZE_BDCFG

/* langpack: this block is used to store the language pack. */
#define LANPACK_OFFSET	(WINDOW_SIZE - SIZE_RGDB - SIZE_LANGPACK)
#define LANPACK_SIZE	SIZE_LANGPACK

/* flash: the whole flash image */
#define FLASH_OFFSET	0
#define FLASH_SIZE		WINDOW_SIZE

/* the partition in block 'bdcfg'. */
#define BOOTCFG_SIZE	0x2000

static struct mtd_partition physmap_partitions[] =
{
	/* The following partitions are the "MUST" in ELBOX. */
	{name:"rootfs",		offset:0x000000,		size:0x000000,		mask_flags:MTD_WRITEABLE, },
	{name:"upgrade",	offset:UPGRADE_OFFSET,	size:UPGRADE_SIZE,	},
	{name:"rgdb",		offset:RGDB_OFFSET,		size:RGDB_SIZE,		},
	{name:"bdcfg",		offset:BDCFG_OFFSET,	size:BDCFG_SIZE,	},
	{name:"langpack",	offset:LANPACK_OFFSET,	size:LANPACK_SIZE,	},
	{name:"flash",		offset:FLASH_OFFSET,	size:FLASH_SIZE,	mask_flags:MTD_WRITEABLE, },

	/* The following partitions are board dependent. */
	{name:"Boot",		offset:0x000000,		size:SIZE_BOOTCODE,	mask_flags:	MTD_WRITEABLE, },
	{name:"Boot Config",offset:BDCFG_OFFSET,	size:BOOTCFG_SIZE,	mask_flags:	MTD_WRITEABLE, },
	{name:NULL,	},
};

/* the tag is 32 bytes octet,
 * first part is the tag string,
 * and the second half is reserved for future used. */
#define PACKIMG_TAG "--PaCkImGs--"
struct packtag
{
	char tag[16];
	unsigned long size;
	char reserved[12];
};

/* Jack add 12/12/07 +++ */
#define IMG_MAX_DEVNAME		32
#define MAX_SIGNATURE		32
typedef struct _imgblock imgblock_t;
struct _imgblock
{
	unsigned long	magic;		/* image magic number (should be IMG_V2_MAGIC_NO in little endian. */
	unsigned long	size;		/* size of the image. */
	unsigned long	offset;		/* offset from the beginning of the storage device. */
	char			devname[IMG_MAX_DEVNAME];	/* null termiated string of the storage device name. ex. "/dev/mtd6" */
	unsigned char	digest[16];	/* MD5 digest of the image */
} __attribute__ ((packed));

typedef struct _imghdr2 imghdr2_t;
struct _imghdr2
{
	char			signature[MAX_SIGNATURE];
	unsigned long	magic;	/* should be IMG_V2_MAGIC_NO in little endian. */
} __attribute__ ((packed));

/* Jack add 12/12/07 --- */
static struct mtd_partition * __init init_mtd_partitions(struct mtd_info * mtd, size_t size)
{
	struct squashfs_super_block *squashfsb;
	struct packtag *ptag = NULL;
	unsigned char buf[512];
	int off = physmap_partitions[1].offset;
    int shift_hdr = 0;  /* Jack add 18/12/07 */
	size_t len;

	squashfsb = (struct squashfs_super_block *) &(buf[32]);

	/* Skip kernel image, and look at every 64KB boundary */
	for (off += CONFIG_ELBOX_KERNEL_SKIP; off < size; off += (64 * 1024))
	{
	    shift_hdr = 0;
		memset(buf, 0xe5, sizeof(buf));

		/* Read block 0 to test for fs superblock */
		if (MTD_READ(mtd, off, sizeof(buf), &len, buf) || len != sizeof(buf))
			continue;

		/* Try to find the tag of packimgs, the tag is always at 64K boundary. */
		if (memcmp(buf, PACKIMG_TAG, 12))
		{
		    /* Jack modify 18/12/07 +++ */
		    // if image header in flash, we should shift (sizeof(imghdr2_t) + sizeof(imgblock_t) ) to find PACKIMG_TAG
		    if (memcmp(buf+(sizeof(imghdr2_t) + sizeof(imgblock_t) ), PACKIMG_TAG, 12))
		    {		        
			continue;
		    }
		    else
		    {   
		        //printk("shift_hdr:%d\n", shift_hdr); 
		        shift_hdr = (sizeof(imghdr2_t) + sizeof(imgblock_t) ) ;
		        squashfsb = (struct squashfs_super_block *) &(buf[32+(sizeof(imghdr2_t) + sizeof(imgblock_t) )]);
		    }
		    //continue;
		    /* Jack modify 18/12/07 --- */
		}
        //printk(" yes, we found it \n"); 
		/* yes, we found it, check for supported file system */

		/* squashfs is at block zero */
		if (squashfsb->s_magic == SQUASHFS_MAGIC)
		{
			printk(KERN_NOTICE
				   "%s: squashfs filesystem found at block %d\n",
				   mtd->name, off/BLOCK_SIZE);
			ptag = (struct packtag *)buf;
			goto done;
		}
	}

	printk(KERN_NOTICE "%s: Couldn't find valid rootfs image!\n", mtd->name);

done:
	if (ptag)
	{
		physmap_partitions[0].offset = off + 32;
		physmap_partitions[0].offset += shift_hdr;  /* Jack add 18/12/07 */
		physmap_partitions[0].size = ntohl(ptag->size);
	}
	else
	{
		physmap_partitions[0].offset = 0;
		physmap_partitions[0].size = 0;
	}

	return physmap_partitions;
}

int __init init_elbox_physmap(void)
{
	int ret;
	struct mtd_partition * parts = 0;
	int i;

	physmap_map.map_priv_1 = (unsigned long)ioremap(WINDOW_ADDR, WINDOW_SIZE);
	physmap_map.map_priv_2 = 0;
	if (!physmap_map.map_priv_1)
	{
		printk("%s: ioremap failed!\n",__FUNCTION__);
		return -EIO;
	}

	if (!(mymtd = do_map_probe("cfi_probe", &physmap_map)))
	{
		printk("%s: cfi_probe failed\n",__FUNCTION__);
		iounmap((void *)physmap_map.map_priv_1);
		physmap_map.map_priv_1 = 0;
		return -ENXIO;
	}

	mymtd->module = THIS_MODULE;
	add_mtd_device(mymtd);
    printk(KERN_NOTICE "Flash device: 0x%x at 0x%x\n", WINDOW_SIZE, WINDOW_ADDR);

	parts = init_mtd_partitions(mymtd, WINDOW_SIZE);
	for (i = 0; parts[i].name; i++);
	ret = add_mtd_partitions(mymtd, parts, i);
	if (ret)
	{
		printk("%s: add_mtd_partitions failed!\n",__FUNCTION__);
		del_mtd_device(mymtd);
		map_destroy(mymtd);
		iounmap((void *)physmap_map.map_priv_1);
		physmap_map.map_priv_1 = 0;
		return ret;
	}

	return 0;
}

static void __exit cleanup_elbox_physmap(void)
{
	if (mymtd)
	{
		del_mtd_partitions(mymtd);
		del_mtd_device(mymtd);
		map_destroy(mymtd);
	}
	if (physmap_map.map_priv_1)
	{
		iounmap((void *)physmap_map.map_priv_1);
		physmap_map.map_priv_1 = 0;
	}
}

module_init(init_elbox_physmap);
module_exit(cleanup_elbox_physmap);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("David Hsieh <david_hsieh@alhpanetworks.com>");
MODULE_DESCRIPTION("ELBOX physmap flash driver.");
