/*
 *	Forwarding decision
 *	Linux ethernet bridge
 *
 *	Authors:
 *	Lennert Buytenhek		<buytenh@gnu.org>
 *
 *	$Id: br_forward.c,v 1.1.1.1 2006-10-13 02:42:51 steven Exp $
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License
 *	as published by the Free Software Foundation; either version
 *	2 of the License, or (at your option) any later version.
 */

#include <linux/kernel.h>
#include <linux/netdevice.h>
#include <linux/inetdevice.h>
#include <linux/skbuff.h>
#include <linux/if_bridge.h>
#include <linux/netfilter_bridge.h>
#include "br_private.h"
#ifdef CONFIG_BRIDGE_IGMPP_PROCFS
#include <linux/ip.h>
#endif
#ifdef CONFIG_BRIDGE_IGMP_SNOOP
#include <linux/ip.h>
#endif

#ifdef CONFIG_BRIDGE_MAT
static void Wlan_Tx(struct sk_buff *skb)
{
    unsigned int                sip=0;
    struct net_bridge           *br;
    struct net_bridge_port      *p;
    struct iphdr                iphd, *iph;
    struct arphdr	            *arphd, arphdh;
    struct postarphdr           *postarphd, postarphdh;

	/*--- Add your filter below ---*/

    p = skb->dev->br_port;
    br = p->br;
    if(ntohs(skb->mac.ethernet->h_proto)==ETH_P_IP){
        /*IP based packets*/
        iph = skb_header_pointer(skb, 0, sizeof(iphd), &iphd);
        memcpy(skb->mac.ethernet->h_source, br->dev.dev_addr, ETH_ALEN);
        memcpy(&sip, &(iph->saddr), sizeof(unsigned int));
#if MAT_DBG
        printk("Wlan_Tx: ip %d.%d.%d.%d(%8x).\n", sip&0x000000ff, sip>>8&0x000000ff, sip>>16&0x000000ff, sip>>24&0x000000ff, sip);
#endif /*MAT_DBG*/
    }else if(ntohs(skb->mac.ethernet->h_proto)==ETH_P_ARP){
        /*ARP packets*/
        arphd = skb_header_pointer(skb, 0, sizeof(arphdh), &arphdh);
        postarphd = skb_header_pointer(skb, sizeof(arphdh), sizeof(postarphdh), &postarphdh);
        memcpy(skb->mac.ethernet->h_source, br->dev.dev_addr, ETH_ALEN);
        memcpy(postarphd->ar_sha, br->dev.dev_addr, ETH_ALEN);
        memcpy(&sip, postarphd->ar_sip, sizeof(unsigned int));
#if MAT_DBG
        printk("Wlan_Tx: arp(%d) %d.%d.%d.%d(%8x).\n", ntohs(arphd->ar_op), sip&0x000000ff, sip>>8&0x000000ff, sip>>16&0x000000ff, sip>>24&0x000000ff, sip);
#endif /*MAT_DBG*/
    }

	return;
}

static void Eth_Tx(struct sk_buff *skb)
{
#if MAT_DBG
    unsigned int                sip=0;
    struct net_bridge           *br;
    struct net_bridge_port      *p;
    struct iphdr                iphd, *iph;
    struct arphdr	            *arphd, arphdh;
    struct postarphdr           *postarphd, postarphdh;
#endif /*MAT_DBG*/

	/*--- Add your filter below ---*/

#if MAT_DBG
    p = skb->dev->br_port;
    br = p->br;
    if(ntohs(skb->mac.ethernet->h_proto)==ETH_P_IP){
        /*IP based packets*/
        iph = skb_header_pointer(skb, 0, sizeof(iphd), &iphd);
        memcpy(&sip, &(iph->saddr), sizeof(unsigned int));
        printk("Eth_Tx: ip %d.%d.%d.%d(%8x).\n", sip&0x000000ff, sip>>8&0x000000ff, sip>>16&0x000000ff, sip>>24&0x000000ff, sip);
    }else if(ntohs(skb->mac.ethernet->h_proto)==ETH_P_ARP){
        /*ARP packets*/
        arphd = skb_header_pointer(skb, 0, sizeof(arphdh), &arphdh);
        postarphd = skb_header_pointer(skb, sizeof(arphdh), sizeof(postarphdh), &postarphdh);
        memcpy(&sip, postarphd->ar_sip, sizeof(unsigned int));
        printk("Eth_Tx: arp(%d) req %d.%d.%d.%d(%8x).\n", ntohs(arphd->ar_op), sip&0x000000ff, sip>>8&0x000000ff, sip>>16&0x000000ff, sip>>24&0x000000ff, sip);
    }
#endif /*MAT_DBG*/

	return;
}

static void frame_Tx(struct sk_buff *skb)
{
	if(!strcmp(skb->dev->name, WLAN)){
		Wlan_Tx(skb);
	}else if(!strcmp(skb->dev->name, ETHERNET)){
		Eth_Tx(skb);
	}

	/*--- Add your filter below ---*/

	return;
}
#endif /*CONFIG_BRIDGE_MAT*/

static inline int should_deliver(struct net_bridge_port *p, struct sk_buff *skb)
{
#ifdef CONFIG_BRIDGE_MULTICAST_BWCTRL
	unsigned char * dest;
#endif

	if (skb->dev == p->dev ||
	    p->state != BR_STATE_FORWARDING)
		return 0;

#ifdef CONFIG_BRIDGE_MULTICAST_BWCTRL
	dest = skb->mac.ethernet->h_dest;
	if ((dest[0] & 1) && p->bandwidth != 0)
	{
		if ((p->accumulation + skb->len) > p->bandwidth) return 0;
		p->accumulation += skb->len;
	}
#endif

	return 1;
}

static int __dev_queue_push_xmit(struct sk_buff *skb)
{
	skb_push(skb, ETH_HLEN);
	dev_queue_xmit(skb);

	return 0;
}

static int __br_forward_finish(struct sk_buff *skb)
{
#ifdef CONFIG_BRIDGE_MAT
struct net_device *nd;
struct net_bridge_port *nbp;
struct net_bridge *br;

nd = skb->dev;
nbp = nd->br_port;
br = nbp->br;
if (br->ap_mode == APC){
	frame_Tx(skb);
}
#endif /*CONFIG_BRIDGE_MAT*/

	NF_HOOK(PF_BRIDGE, NF_BR_POST_ROUTING, skb, NULL, skb->dev,
			__dev_queue_push_xmit);

	return 0;
}

static void __br_deliver(struct net_bridge_port *to, struct sk_buff *skb)
{
	skb->dev = to->dev;
	NF_HOOK(PF_BRIDGE, NF_BR_LOCAL_OUT, skb, NULL, skb->dev,
			__br_forward_finish);
}

static void __br_forward(struct net_bridge_port *to, struct sk_buff *skb)
{
	struct net_device *indev;

	indev = skb->dev;
	skb->dev = to->dev;
	skb->ip_summed = CHECKSUM_NONE;

	NF_HOOK(PF_BRIDGE, NF_BR_FORWARD, skb, indev, skb->dev,
			__br_forward_finish);
}

/* called under bridge lock */
void br_deliver(struct net_bridge_port *to, struct sk_buff *skb)
{
	if (should_deliver(to, skb)) {
		__br_deliver(to, skb);
		return;
	}

	kfree_skb(skb);
}

/* called under bridge lock */
void br_forward(struct net_bridge_port *to, struct sk_buff *skb)
{
	if (should_deliver(to, skb)) {
		__br_forward(to, skb);
		return;
	}

	kfree_skb(skb);
}

#ifdef CONFIG_BRIDGE_IGMP_SNOOP
static void copy_ap_mac(unsigned char* to, unsigned char * from)
{
	int i;
	for(i=0; i<6; i++)
		*(to+i)=*(from+i);
	return;
}
#endif //CONFIG_BRIDGE_IGMP_SNOOP

#ifdef CONFIG_BRIDGE_IGMPP_PROCFS
static void copy_mac(unsigned char* to, unsigned char * from)
{
	int i;
	for(i=0; i<6; i++)
		*(to+i)=*(from+i);
	return;
}
#endif

/* called under bridge lock */
static void br_flood(struct net_bridge *br, struct sk_buff *skb, int clone,
	void (*__packet_hook)(struct net_bridge_port *p, struct sk_buff *skb))
{
	struct net_bridge_port *p;
	struct net_bridge_port *prev;
	unsigned char *dest;
	char system_on_subent[6]={0x01,0x00,0x5e,0x00,0x00,0x01};
	char routers_on_subent[6]={0x01,0x00,0x5e,0x00,0x00,0x02};
	dest = skb->mac.ethernet->h_dest;
	if (clone) {
		struct sk_buff *skb2;

		if ((skb2 = skb_clone(skb, GFP_ATOMIC)) == NULL) {
			br->statistics.tx_dropped++;
			return;
		}

		skb = skb2;
	}
#ifdef CONFIG_BRIDGE_IGMP_SNOOP
if(br->igmp_enable==1){
	if(atomic_read(&br->br_igmp_ap_table_enable) == 1){
		p = br->port_list;
		while (p != NULL) {
			struct sk_buff *skb2;
			if ( atomic_read(&p->ap_wireless_interface) == 1){
				read_lock(&p->port_igmp_ap_table_rwlock); //reader spin lock
				int groupIdx;
				groupIdx = search_group_ap_IP(&p->port_igmp_ap_table, skb->nh.iph->daddr);
				
				if (groupIdx >=0){
					/* skb_copy for each host*/
					int i;
					for(i=0; i<HOSTLIST_NUMBER; i++){	
						if (p->port_igmp_ap_table.group_list[groupIdx].host_list[i].used ==1){
							if ((skb2 = skb_copy(skb, GFP_ATOMIC)) == NULL) {
								br->statistics.tx_dropped++;
								kfree_skb(skb);
								read_unlock(&p->port_igmp_ap_table_rwlock); // reader spin unlock
								return;
							}
							copy_ap_mac(skb2->mac.ethernet->h_dest, 
									p->port_igmp_ap_table.group_list[groupIdx].host_list[i].mac_addr);
							if (should_deliver(p, skb2))
								__packet_hook(p, skb2);
							else
								kfree_skb(skb2);
						}// if used - END
					}// for loop - END
				}else {											
						if(dest[0] & 1){
							if(dest[0]==0xff){
								/* skb's destination IP address is not stored in port_igmp_ap_table, 
								 * so should be deliver/forward it anwary. 
								 * skb_clone..... */								 
								if ((skb2 = skb_clone(skb, GFP_ATOMIC)) == NULL) {
									br->statistics.tx_dropped++;
									kfree_skb(skb);
									read_unlock(&p->port_igmp_ap_table_rwlock); // reader spin unlock
									return;
								}
								if (should_deliver(p,skb2))
									__packet_hook(p, skb2);
								else
									kfree_skb(skb2);
									}
							else{
								if(!memcmp(dest,system_on_subent,6)||!memcmp(dest,routers_on_subent,6))
									{									
									if ((skb2 = skb_clone(skb, GFP_ATOMIC)) == NULL) {
										br->statistics.tx_dropped++;
										kfree_skb(skb);
										read_unlock(&p->port_igmp_ap_table_rwlock); // reader spin unlock
										return;
										}
									if (should_deliver(p,skb2))
										__packet_hook(p, skb2);
									else
										kfree_skb(skb2);
							     	}						
							   else {								
									kfree_skb(skb);								
									return;
							     	}
							    }
					}
				}// groupIdx >=0 - END
				
				read_unlock(&p->port_igmp_ap_table_rwlock); // reader spin unlock
			}else{ 
				/* it's wired interface */
				/* skb_clone..... */
				if ((skb2 = skb_clone(skb, GFP_ATOMIC)) == NULL) {
					br->statistics.tx_dropped++;
					kfree_skb(skb);
					return;
				}
				if (should_deliver(p,skb2))
					__packet_hook(p, skb2);
				else
					kfree_skb(skb2);

			}// interface ==1 - END
			p = p->next;
		} //while loop - END
		kfree_skb(skb);

	  }
	}else{

		prev = NULL;
		p = br->port_list;
		while (p != NULL) {
			if (should_deliver(p, skb)) {
				if (prev != NULL) {
					struct sk_buff *skb2;
					if ((skb2 = skb_clone(skb, GFP_ATOMIC)) == NULL) {
						br->statistics.tx_dropped++;
						kfree_skb(skb);
						return;
					}
					__packet_hook(prev, skb2);
				}
				prev = p;
			}
			p = p->next;
		}
		if (prev != NULL) {
			__packet_hook(prev, skb);
			return;
		}
		kfree_skb(skb);
	}
#else
#ifdef CONFIG_BRIDGE_IGMPP_PROCFS
	if(atomic_read(&br->br_igmpp_table_enable) == 1){
		p = br->port_list;
		while (p != NULL) {
			struct sk_buff *skb2;
			if ( atomic_read(&p->wireless_interface) == 1){
				read_lock(&p->port_igmpp_table_rwlock); //reader spin lock
				int groupIdx ;
				groupIdx = search_group_IP( &p->port_igmpp_table, skb->nh.iph->daddr);
				if (groupIdx >=0){
					/* skb_copy for each host*/
					int i;
					for(i=0; i<HOSTLIST_NUMBER; i++){	
						if (p->port_igmpp_table.group_list[groupIdx].host_list[i].used ==1){
							if ((skb2 = skb_copy(skb, GFP_ATOMIC)) == NULL) {
								br->statistics.tx_dropped++;
								kfree_skb(skb);
								read_unlock(&p->port_igmpp_table_rwlock); // reader spin unlock
								return;
							}
							copy_mac(skb2->mac.ethernet->h_dest, 
									p->port_igmpp_table.group_list[groupIdx].host_list[i].mac_addr);
							if (should_deliver(p, skb2))
								__packet_hook(p, skb2);
							else
								kfree_skb(skb2);
						}// if used - END
					}// for loop - END
				}else {
					/* skb's destination IP address is not stored in port_igmpp_table, 
					 * so should be deliver/forward it anwary. 
					 * skb_clone..... */
					if ((skb2 = skb_clone(skb, GFP_ATOMIC)) == NULL) {
						br->statistics.tx_dropped++;
						kfree_skb(skb);
						read_unlock(&p->port_igmpp_table_rwlock); // reader spin unlock
						return;
					}
					if (should_deliver(p,skb2))
						__packet_hook(p, skb2);
					else
						kfree_skb(skb2);
				}// groupIdx >=0 - END
				read_unlock(&p->port_igmpp_table_rwlock); // reader spin unlock
			}else{ 
				/* it's wired interface */
				/* skb_clone..... */
				if ((skb2 = skb_clone(skb, GFP_ATOMIC)) == NULL) {
					br->statistics.tx_dropped++;
					kfree_skb(skb);
					return;
				}
				if (should_deliver(p,skb2))
					__packet_hook(p, skb2);
				else
					kfree_skb(skb2);

			}// interface ==1 - END
			p = p->next;
		} //while loop - END
		kfree_skb(skb);

	}else{
		prev = NULL;
		p = br->port_list;
		while (p != NULL) {
			if (should_deliver(p, skb)) {
				if (prev != NULL) {
					struct sk_buff *skb2;
					if ((skb2 = skb_clone(skb, GFP_ATOMIC)) == NULL) {
						br->statistics.tx_dropped++;
						kfree_skb(skb);
						return;
					}
					__packet_hook(prev, skb2);
				}
				prev = p;
			}
			p = p->next;
		}
		if (prev != NULL) {
			__packet_hook(prev, skb);
			return;
		}
		kfree_skb(skb);
	}
#else

	prev = NULL;

	p = br->port_list;
	while (p != NULL) {
		if (should_deliver(p, skb)) {
			if (prev != NULL) {
				struct sk_buff *skb2;

				if ((skb2 = skb_clone(skb, GFP_ATOMIC)) == NULL) {
					br->statistics.tx_dropped++;
					kfree_skb(skb);
					return;
				}

				__packet_hook(prev, skb2);
			}

			prev = p;
		}

		p = p->next;
	}

	if (prev != NULL) {
		__packet_hook(prev, skb);
		return;
	}

	kfree_skb(skb);
#endif
#endif //CONFIG_BRIDGE_IGMP_SNOOP
}

/* called under bridge lock */
void br_flood_deliver(struct net_bridge *br, struct sk_buff *skb, int clone)
{
	br_flood(br, skb, clone, __br_deliver);
}

/* called under bridge lock */
void br_flood_forward(struct net_bridge *br, struct sk_buff *skb, int clone)
{
	br_flood(br, skb, clone, __br_forward);
}
