/*
 *	Userspace interface
 *	Linux ethernet bridge
 *
 *	Authors:
 *	Lennert Buytenhek		<buytenh@gnu.org>
 *
 *	$Id: br_if.c,v 1.1.1.1 2006-10-13 02:42:51 steven Exp $
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License
 *	as published by the Free Software Foundation; either version
 *	2 of the License, or (at your option) any later version.
 */

#include <linux/kernel.h>
#include <linux/if_arp.h>
#include <linux/if_bridge.h>
#include <linux/inetdevice.h>
#include <linux/rtnetlink.h>
#include <linux/brlock.h>
#include <linux/etherdevice.h>
#include <asm/uaccess.h>
#ifdef CONFIG_BRIDGE_FORWARD_CTRL
#include <linux/proc_fs.h>
#endif
#include "br_private.h"
#ifdef CONFIG_BRIDGE_IGMP_SNOOP
#include <linux/proc_fs.h>
#endif
static int br_initial_port_cost(struct net_device *dev)
{
	if (!strncmp(dev->name, "lec", 3))
		return 7;

	if (!strncmp(dev->name, "eth", 3))
		return 100;			/* FIXME handle 100Mbps */

	if (!strncmp(dev->name, "plip", 4))
		return 2500;

	return 100;
}

/* called under BR_NETPROTO_LOCK and bridge lock */
static int __br_del_if(struct net_bridge *br, struct net_device *dev)
{
	struct net_bridge_port *p;
	struct net_bridge_port **pptr;

	if ((p = dev->br_port) == NULL || p->br != br)
		return -EINVAL;

	br_stp_disable_port(p);

	dev_set_promiscuity(dev, -1);
	dev->br_port = NULL;

	pptr = &br->port_list;
	while (*pptr != NULL) {
		if (*pptr == p) {
			*pptr = p->next;
			break;
		}

		pptr = &((*pptr)->next);
	}

	br_fdb_delete_by_port(br, p);
	kfree(p);
	dev_put(dev);

	return 0;
}

static void del_ifs(struct net_bridge *br)
{
	br_write_lock_bh(BR_NETPROTO_LOCK);
	write_lock(&br->lock);
	while (br->port_list != NULL)
		__br_del_if(br, br->port_list->dev);
	write_unlock(&br->lock);
	br_write_unlock_bh(BR_NETPROTO_LOCK);
}

static struct net_bridge *new_nb(char *name)
{
	struct net_bridge *br;
	struct net_device *dev;

	if ((br = kmalloc(sizeof(*br), GFP_KERNEL)) == NULL)
		return NULL;

	memset(br, 0, sizeof(*br));
	dev = &br->dev;

	strncpy(dev->name, name, IFNAMSIZ);
	dev->priv = br;
	ether_setup(dev);
	br_dev_setup(dev);

	br->lock = RW_LOCK_UNLOCKED;
	br->hash_lock = RW_LOCK_UNLOCKED;

	br->bridge_id.prio[0] = 0x80;
	br->bridge_id.prio[1] = 0x00;
	memset(br->bridge_id.addr, 0, ETH_ALEN);

	br->stp_enabled = 0;
	br->designated_root = br->bridge_id;
	br->root_path_cost = 0;
	br->root_port = 0;
	br->bridge_max_age = br->max_age = 20 * HZ;
	br->bridge_hello_time = br->hello_time = 2 * HZ;
	br->bridge_forward_delay = br->forward_delay = 15 * HZ;
	br->topology_change = 0;
	br->topology_change_detected = 0;
	br_timer_clear(&br->hello_timer);
	br_timer_clear(&br->tcn_timer);
	br_timer_clear(&br->topology_change_timer);

	br->ageing_time = 300 * HZ;
	br->gc_interval = 4 * HZ;
#ifdef CONFIG_BRIDGE_IGMP_SNOOP
	br->igmp_switch_query_timer =255;
#endif
	return br;
}

/* called under bridge lock */
static struct net_bridge_port *new_nbp(struct net_bridge *br, struct net_device *dev)
{
	int i;
	struct net_bridge_port *p;

	p = kmalloc(sizeof(*p), GFP_ATOMIC);
	if (p == NULL)
		return p;

	memset(p, 0, sizeof(*p));
	p->br = br;
	p->dev = dev;
	p->path_cost = br_initial_port_cost(dev);
	p->priority = 0x80;

	for (i=1;i<255;i++)
		if (br_get_port(br, i) == NULL)
			break;

	if (i == 255) {
		kfree(p);
		return NULL;
	}

	dev->br_port = p;

	p->port_no = i;
	br_init_port(p);
	p->state = BR_STATE_DISABLED;

	p->next = br->port_list;
	br->port_list = p;

	return p;
}
#ifdef CONFIG_BRIDGE_IGMP_SNOOP
void trans_32to8(uint32_t *ip, uint8_t **a)
{
	*a = (uint8_t *)ip;
	return;
}

/* If host MAC has been existed in group list, return index of host MAC list(positive numbe),
 * else return -1.
 */
static int proc_read_br_igmp_ap (char *buf, char **start, off_t offset,
		int len, int *eof, void *data)
{
	int count =0;
	struct net_bridge *br = (struct net_bridge *) data;

	printk( "**************************************\n");
	printk( "* bridge name: %s\n",br->dev.name);
	printk( "* br_igmpp_table_enable-> %d\n",atomic_read(&br->br_igmp_ap_table_enable));
	printk( "* * * * * * * * * * * * * * * * *\n");
	struct net_bridge_port *p ;
	p = br->port_list;
	while ( p!=NULL){
		printk( "* <%d> port name : %s\n", p->port_no, p->dev->name);
		printk( "* <%d> table size: %u x %u\n", p->port_no, 30, HOSTLIST_NUMBER );
		printk( "* <%d> ap_wireless_interface -> %d\n", p->port_no, atomic_read(&p->ap_wireless_interface) );
		read_lock(&p->port_igmp_ap_table_rwlock); //reader spin lock
		printk( "* <%d> port_igmp_ap_table -> %d\n", p->port_no, p->port_igmp_ap_table.enable );
		printk( "* ===============================\n");
		int i,j,k;
		for( i=0; i<30; i++){
			if(p->port_igmp_ap_table.group_list[i].used == 1){
				uint32_t ip32_addr = p->port_igmp_ap_table.group_list[i].ip_addr;
				uint8_t *ip8_addr;
				trans_32to8(&ip32_addr, &ip8_addr);	
				printk( "* <%d> => group_list[%u]= %u %u %u %u\n", p->port_no, 
						i, *ip8_addr, *(ip8_addr+1), *(ip8_addr+2), *(ip8_addr+3));
				printk( "* <%d> agint time: %u sec\n", p->port_no, p->port_igmp_ap_table.group_list[i].port_aging_time );
				printk( "* <%d> flag: %u \n", p->port_no, p->port_igmp_ap_table.group_list[i].in_special_query );
				printk( "* <%d> response time: %u sec\n", p->port_no, p->port_igmp_ap_table.group_list[i].max_response_time );
				for (j=0; j<HOSTLIST_NUMBER; j++){
					if(p->port_igmp_ap_table.group_list[i].host_list[j].used == 1){
						printk("* <%d> -", p->port_no);
						for (k=0; k<6; k++)
							printk( " %02X ", p->port_igmp_ap_table.group_list[i].host_list[j].mac_addr[k] ); 
						printk("\n");
					}
				}
				printk( "*     - - - - - - - - - - - - -\n");
			}
		}
		read_unlock(&p->port_igmp_ap_table_rwlock); // reader spin unlock
		p=p->next;
	}
	printk( "**************************************\n");

	*eof = 1;
	return count;
}






/* For different platform, 
 * kernel API may don't support strcmp() originally, 
 * we copy strcmp() and named strcmp1() for conveniently.
 *
 * strcmp - Compare two strings
 * @cs: One string
 * @ct: Another string
 */
static int ap_strcmp1(char * cs,char * ct)
{
	register signed char __res;
	while (1) {
		if ((__res = *cs - *ct++) != 0 || !*cs++)
			break;
	}
	return __res;
}




static struct net_bridge_port * search_ap_device(struct net_bridge * br, char* name)
{
	struct net_bridge_port *p;

	p = br->port_list;
	while(p != NULL){

		if (ap_strcmp1(p->dev->name, name) == 0 ){
			return p;
		}
		p = p->next;
	}
	return NULL;
}

/* enable each port's port_igmp_ap_table */
static void table_all_enable(struct net_bridge *br)
{
	/* set bridge */
	atomic_set(&br->br_igmp_ap_table_enable, 1);
	/* set each port */
	struct net_bridge_port *p;
	p = br->port_list;
	while( p != NULL){
		write_lock(&p->port_igmp_ap_table_rwlock);
		memset( &p->port_igmp_ap_table, 0, sizeof( struct port_igmp_ap_table_t));
		p->port_igmp_ap_table.enable = 1;
		write_unlock(&p->port_igmp_ap_table_rwlock);
		p=p->next;
	}
}
/* disable each port's port_igmp_ap_table */
static void table_all_disable(struct net_bridge *br)
{
	/* set bridge */
	atomic_set(&br->br_igmp_ap_table_enable, 0);
	/* set each port */
	struct net_bridge_port *p;
	p = br->port_list;
	while( p != NULL){
		write_lock(&p->port_igmp_ap_table_rwlock);
		p->port_igmp_ap_table.enable = 0;
		write_unlock(&p->port_igmp_ap_table_rwlock);
		p=p->next;
	}
}
static int table_set_ap_wl(struct net_bridge *br, char * name)
{
	struct net_bridge_port *hit_port;
	hit_port = search_ap_device(br, name);
	if (hit_port != NULL){
		atomic_set(&hit_port->ap_wireless_interface, 1);
		return 0;
	}else 
		return (-1);
}
static int table_ap_unsetwl(struct net_bridge *br, char * name)
{
	struct net_bridge_port *hit_port;
	hit_port = search_ap_device(br, name);
	if (hit_port != NULL){
		atomic_set(&hit_port->ap_wireless_interface, 0);
		return 0;
	}else 
		return (-1);
}
static int proc_write_br_igmp_ap (struct file *file, const char *buf,
		unsigned long count, void *data)
{
	int len;
	if(count > MESSAGE_LENGTH) {len = MESSAGE_LENGTH;}
	else {len = count; }
	char message[len];
	if(copy_from_user(message, buf, len))
		return -EFAULT;
	message[len-1] = '\0';

	/* split input message that get from user space 
	 * token[0] => action token --> add or remove 
	 * token[1] => multicast group IP address 
	 * token[2] => MAC address of host IP  
	 */
	char *msgDelim = MESSAGE_DELIM;
	char *pmesg = message ;
	char *action_token, *action;

	action_token = strsep(&pmesg, msgDelim);

	struct net_bridge *br = (struct net_bridge *) data;

	/* ============================  enable each port's port_igmp_ap_table ====================*/
	/* NOTE: Each port_igmp_ap_table will be clean.								 			 */
	action = ACTION_ENABLE_TABLE;
	if ( memcmp(action_token, action, sizeof(ACTION_ENABLE_TABLE) )== 0){
		table_all_enable(br);
		printk(KERN_INFO "[BR_IGMP_AP_PROC]->ENABLE %s\n",br->dev.name);
		
	}

	/* ============================  disable each port's port_igmp_ap_table ===================*/
	action = ACTION_DISABLE_TABLE;
	if ( memcmp(action_token, action, sizeof(ACTION_DISABLE_TABLE) )== 0){
		table_all_disable(br);
		printk(KERN_INFO "[BR_IGMP_AP_PROC]->DISABLE %s\n",br->dev.name);
		
	}

	/* ============================  set wireless ============================*/
	action = ACTION_SET_WL;
	if ( memcmp(action_token, action, sizeof(ACTION_SET_WL) )== 0){
		if (table_set_ap_wl(br,pmesg) == 0){
			printk(KERN_INFO "[BR_IGMP_AP_PROC]->SETWL-> %s \n",pmesg);
			
		}else{
			printk(KERN_INFO "[BR_IGMP_AP_PROC]->WARNING SETWL FAILURE-> %s\n",pmesg);
			goto proc_write_br_igmpp_out;
		}
	}
	
	/* ============================  unset wireless ==========================*/
	action = ACTION_UNSET_WL;
	if ( memcmp(action_token, action, sizeof(ACTION_UNSET_WL) )== 0){
		if (table_ap_unsetwl(br,pmesg) == 0){
			printk(KERN_INFO "[BR_IGMP_AP_PROC]->UNSETWL-> %s \n",pmesg);
			
		}else{
			printk(KERN_INFO "[BR_IGMP_AP_PROC]->WARNING UNSETWL FAILURE-> %s\n",pmesg);
			goto proc_write_br_igmpp_out;
		}
	}	
proc_write_br_igmpp_out:
	return len;

}
static int proc_write_br_query (struct file *file, const char *buf,
		unsigned long count, void *data)
{
}


static int proc_read_br_query (char *buf, char **start, off_t offset,
		int len, int *eof, void *data)
{
	struct net_bridge *br = (struct net_bridge *) data;
	struct net_device dev = br->dev;
	char * p = buf;
	u_int32_t br_change_ip = 0x00000000;
	int i,j,k,l;		
	 i = br->dev.br_ip & 0x000000ff;
         i = i<<24;
         j = br->dev.br_ip & 0x0000ff00;
         j = j << 8;
         k = br->dev.br_ip & 0x00ff0000;
         k = k >> 8;
         l = br->dev.br_ip & 0xff000000;
         l = l >>24;
         br_change_ip =(i|j|k|l);	
	
	p += sprintf(p, "%d",atomic_read(&br->br_igmp_AP_query_enable));
    p += sprintf(p,"%x",br_change_ip);
	*eof = 1;	
	return p - buf;
}

/* procfs write/read for br_mac_table */
#if 0
static int proc_read_br_ap_mac (char *buf, char **start, off_t offset,
		int len, int *eof, void *data)
{

	int count =0;
	struct net_bridge *br = (struct net_bridge *) data;
	struct br_mac_ap_table_t *tlist;

	read_lock(&br->br_mac_table_ap_rwlock); //reader spin lock

	printk( "*********************************\n");
	printk( "* br->br_mac_ap_table_enable : %d\n",atomic_read(&br->br_mac_ap_table_enable));
	printk( "*********************************\n");
	printk( "IP address         MAC           \n");

	list_for_each_entry(tlist, &(br->br_mac_ap_table.list), list){
		uint32_t ip32_addr = tlist->ip_addr;
		uint8_t *ip8_addr;
		trans_32to8(&ip32_addr, &ip8_addr);
		printk("%u.%u.%u.%u \t %X:%X:%X:%X:%X:%X \n", 
			*ip8_addr, *(ip8_addr+1), *(ip8_addr+2), *(ip8_addr+3),
			tlist->mac_addr[0],	tlist->mac_addr[1], tlist->mac_addr[2], tlist->mac_addr[3],	
			tlist->mac_addr[4], tlist->mac_addr[5]
		);
	}

	read_unlock(&br->br_mac_table_ap_rwlock); // reader spin unlock

	*eof = 1;
	return count;
}


static int proc_write_br_ap_mac (struct file *file, const char *buf,
		unsigned long count, void *data)
{

	int len;
	if(count > MESSAGE_LENGTH) {len = MESSAGE_LENGTH;}
	else {len = count; }
	char message[len];
	if(copy_from_user(message, buf, len))
		return -EFAULT;
	message[len-1] = '\0';

	char *action;
	struct net_bridge *br = (struct net_bridge *) data;
	write_lock(&br->br_mac_table_ap_rwlock); // writer spin lock

	/* ============================  enable br_mac_table ===================*/
	/* NOTE: The br_mac_table will be clean.								  */
	action = ACTION_ENABLE_TABLE;
	if ( memcmp(message, action, sizeof(ACTION_ENABLE_TABLE) )== 0){
		/* reset br_mac_table */
		struct br_mac_ap_table_t *tlist;
		struct list_head *pos, *q;
		list_for_each_safe(pos, q, &(br->br_mac_ap_table.list) ){
			tlist= list_entry(pos, struct br_mac_ap_table_t, list);
			list_del(pos);
			kfree(tlist);
		}
		memset( &(br->br_mac_ap_table), 0, sizeof( struct br_mac_ap_table_t));
		INIT_LIST_HEAD(&br->br_mac_ap_table.list);
		atomic_set(&br->br_mac_ap_table_enable, 1);
	
		printk(KERN_INFO "[BR_MAC_AP_PROC]->ENABLE %s\n", br->dev.name);
		printk(KERN_INFO "br_mac_ap_table_enable = %d\n", atomic_read(&br->br_mac_ap_table_enable) );
	
	}

	/* ============================  disable br_mac_table ===================*/
	action = ACTION_DISABLE_TABLE;
	if ( memcmp(message, action, sizeof(ACTION_DISABLE_TABLE) )== 0){
		atomic_set(&br->br_mac_ap_table_enable, 0);

		printk(KERN_INFO "[BR_MAC_AP_PROC]->DISABLE %s\n", br->dev.name);
		printk(KERN_INFO "br_mac_ap_table_enable = %d\n", atomic_read(&br->br_mac_ap_table_enable) );

	}

	write_unlock(&br->br_mac_table_ap_rwlock); // writer spin unlock
	return len;
}
#endif
#endif //CONFIG_BRIDGE_IGMP_SNOOP


#ifdef CONFIG_BRIDGE_IGMPP_PROCFS
void trans_32to8(uint32_t *ip, uint8_t **a)
{
	*a = (uint8_t *)ip;
	return;
}

/* If host MAC has been existed in group list, return index of host MAC list(positive numbe),
 * else return -1.
 */
static int search_list_MAC(struct port_igmpp_table_t *pt, int groupIndex, unsigned char* mac_addr)
{
    int i;
	for (i=0; i<HOSTLIST_NUMBER; ++i){
		if (pt->group_list[groupIndex].host_list[i].used==1)
			if (pt->group_list[groupIndex].host_list[i].mac_addr[0] == *mac_addr				)
			  if (pt->group_list[groupIndex].host_list[i].mac_addr[1] == *(mac_addr+1)			)
			    if (pt->group_list[groupIndex].host_list[i].mac_addr[2] == *(mac_addr+2)		)
			      if (pt->group_list[groupIndex].host_list[i].mac_addr[3] == *(mac_addr+3)		)
			        if (pt->group_list[groupIndex].host_list[i].mac_addr[4] == *(mac_addr+4)	)
			         if (pt->group_list[groupIndex].host_list[i].mac_addr[5] == *(mac_addr+5)	)
						return i;
	}
	return (-1);
}
/* If mulsticast group ip has been existed in table, return index of group list(positive numbe),
 * else return -1.
 */
int search_group_IP(struct port_igmpp_table_t *pt, uint32_t ip_addr)
{
    int i;
    for (i=0; i< GROUPLIST_NUMBER; ++i){
        if (pt->group_list[i].used == 1 )
            if (pt->group_list[i].ip_addr == ip_addr)
				return i;
    }
    return (-1);
}
static int get_element_from_MAC_pool(struct port_igmpp_table_t *pt ,int gIdx)
{
    int i;
    for(i=0; i<HOSTLIST_NUMBER; ++i){
        if (pt->group_list[gIdx].host_list[i].used == 0) return i;
    }
    return (-1);
}
static uint8_t * hex2dec(char *ch)
{
	if ( *ch == '0') *ch=0;
	if ( *ch == '1') *ch=1;
	if ( *ch == '2') *ch=2;
	if ( *ch == '3') *ch=3;
	if ( *ch == '4') *ch=4;
	if ( *ch == '5') *ch=5;
	if ( *ch == '6') *ch=6;
	if ( *ch == '7') *ch=7;
	if ( *ch == '8') *ch=8;
	if ( *ch == '9') *ch=9;
	if ( *ch == 'A' || *ch == 'a')	*ch=10;
	if ( *ch == 'B' || *ch == 'b')	*ch=11;
	if ( *ch == 'C' || *ch == 'c') 	*ch=12;
	if ( *ch == 'D' || *ch == 'd')	*ch=13;
	if ( *ch == 'E' || *ch == 'e')	*ch=14;
	if ( *ch == 'F' || *ch == 'f')	*ch=15;
	return ch;
}

static int add_MAC_2_pool(struct port_igmpp_table_t *pt, int gIdx, int mIdx, unsigned char* mac_addr)
{
	pt->group_list[gIdx].host_list[mIdx].used =1;
	int i;
	for (i=0; i<6; i++)
		pt->group_list[gIdx].host_list[mIdx].mac_addr[i] = *(mac_addr+i);
	return 0;
}
static int check_GROUP_pool(struct port_igmpp_table_t *pt)
{
	int i;
	for( i=0; i<GROUPLIST_NUMBER; ++i){
		if( pt->group_list[i].used == 0) return i;
	}
	return -1;
}

static int add_GROUP(struct port_igmpp_table_t *pt, int gIdx, uint32_t ip_addr)
{
	pt->group_list[gIdx].used = 1;
	pt->group_list[gIdx].ip_addr = ip_addr;
	return 0;
}
static int remove_MAC_from_pool(struct port_igmpp_table_t *pt, int gIdx, int mIdx)
{
	pt->group_list[gIdx].host_list[mIdx].used = 0;
	int i;
	for (i=0; i<6; i++)
		pt->group_list[gIdx].host_list[mIdx].mac_addr[i] = 0;
	return 0;
}

static int remove_GROUP_from_pool(struct port_igmpp_table_t *pt, int gIdx)
{
	pt->group_list[gIdx].used = 0;
	pt->group_list[gIdx].ip_addr = 0;
	return 0;
}
static int check_GROUP_is_empty_and_remove(struct port_igmpp_table_t *pt, int gIdx)
{
	int i,used=0;
	for (i=0; i<HOSTLIST_NUMBER; ++i){ 
		if (pt->group_list[gIdx].host_list[i].used == 1)
			++used;
	}
	if (used == 0){
		#ifdef CONFIG_BRIDGE_IGMPP_PROCFS_DEBUG
		uint32_t ip32_addr = pt->group_list[gIdx].ip_addr;
		uint8_t *ip8_addr;
		trans_32to8(&ip32_addr, &ip8_addr);
		printk(KERN_INFO "[BR_IGMPP_PROC]-> Group IP: %u.%u.%u.%u remove !!\n",
													*ip8_addr, *(ip8_addr+1), *(ip8_addr+2), *(ip8_addr+3));
		#endif
		remove_GROUP_from_pool(pt, gIdx);	
		return 0;
	}
	else 
		return 0;
}

static int proc_read_br_igmpp (char *buf, char **start, off_t offset,
		int len, int *eof, void *data)
{
	int count =0;
	struct net_bridge *br = (struct net_bridge *) data;

	printk( "**************************************\n");
	printk( "* bridge name: %s\n",br->dev.name);
	printk( "* br_igmpp_table_enable-> %d\n",atomic_read(&br->br_igmpp_table_enable));
	printk( "* * * * * * * * * * * * * * * * *\n");
	struct net_bridge_port *p ;
	p = br->port_list;
	while ( p!=NULL){
		printk( "* <%d> port name : %s\n", p->port_no, p->dev->name);
		printk( "* <%d> table size: %u x %u\n", p->port_no, GROUPLIST_NUMBER, HOSTLIST_NUMBER );
		printk( "* <%d> wireless_interface -> %d\n", p->port_no, atomic_read(&p->wireless_interface) );
		read_lock(&p->port_igmpp_table_rwlock); //reader spin lock
		printk( "* <%d> port_igmpp_table -> %d\n", p->port_no, p->port_igmpp_table.enable );
		printk( "* ===============================\n");
		int i,j,k;
		for( i=0; i<GROUPLIST_NUMBER; i++){
			if(p->port_igmpp_table.group_list[i].used == 1){
				uint32_t ip32_addr = p->port_igmpp_table.group_list[i].ip_addr;
				uint8_t *ip8_addr;
				trans_32to8(&ip32_addr, &ip8_addr);	
				printk( "* <%d> => group_list[%u]= %u %u %u %u\n", p->port_no, 
						i, *ip8_addr, *(ip8_addr+1), *(ip8_addr+2), *(ip8_addr+3));
				for (j=0; j<HOSTLIST_NUMBER; j++){
					if(p->port_igmpp_table.group_list[i].host_list[j].used == 1){
						printk("* <%d> -", p->port_no);
						for (k=0; k<6; k++)
							printk( " %02X ", p->port_igmpp_table.group_list[i].host_list[j].mac_addr[k] ); 
						printk("\n");
					}
				}
				printk( "*     - - - - - - - - - - - - -\n");
			}
		}
		read_unlock(&p->port_igmpp_table_rwlock); // reader spin unlock
		p=p->next;
	}
	printk( "**************************************\n");

	*eof = 1;
	return count;
}
static void split_IP(uint32_t *ip32_addr, char * token)
{
	char *ipDelim = IP_DELIM;
	char **pIP = &token;
	char *ipField_char[4];
	uint8_t ipField_int[4];
	int i;
	for (i=0; i<4; ++i)
	{
		ipField_char[i] = strsep(pIP, ipDelim);
		ipField_int[i] = (uint8_t) simple_strtoul( ipField_char[i], NULL, 10);
	}
	uint8_t *ip8_addr;
	trans_32to8(ip32_addr, &ip8_addr);	
	for (i =0; i<4; ++i)
		*(ip8_addr+i) = ipField_int[i];
	return;
}
static void split_MAC(unsigned char * mac_addr, char * token)
{
	char *macDelim = MAC_DELIM;
	char **pMAC = &token;
	char * macField_char[6];
	int i;
	for (i=0; i<6; ++i)
	{
		macField_char[i] = strsep(pMAC, macDelim);
		int j;
		/* copy each char byte and convert to dec number */
		char temp[2];
		for(j=0; j<2; ++j) { 
			memcpy(&temp[j],macField_char[i]+j, sizeof(char));
			hex2dec(&temp[j]);
		}

		temp[0] = temp[0] << 4;
		*(mac_addr + i)= (temp[0]^temp[1]);
	}
}

static void table_add(	struct net_bridge_port *p,	struct port_igmpp_table_t *pt,
						uint32_t ip32_addr, 	unsigned char * mac_addr)
{
	/* search group IP */
	int groupIdx,ipIdx;
	groupIdx = search_group_IP(pt, ip32_addr);
	if(groupIdx >= 0){
		/* search list MAC */
		ipIdx = search_list_MAC(pt, groupIdx, mac_addr);
		
		if(ipIdx >= 0){
			#ifdef CONFIG_BRIDGE_IGMPP_PROCFS_DEBUG
			printk(KERN_INFO "[BR_IGMPP_PROC]-> MAC: %X:%X:%X:%X:%X:%X has been existed !!\n",
						*mac_addr,*(mac_addr+1),*(mac_addr+2),*(mac_addr+3),*(mac_addr+4),*(mac_addr+5) );
			#endif
		}else{
			/* check MAC pool */
			int macPoolIdx;
			macPoolIdx = get_element_from_MAC_pool(pt, groupIdx);
			if (macPoolIdx >=0 ){
				/* add MAC to pool */
				add_MAC_2_pool(pt, groupIdx, macPoolIdx, mac_addr);
				#ifdef CONFIG_BRIDGE_IGMPP_PROCFS_DEBUG
				printk(KERN_INFO "[BR_IGMPP_PROC]-> MAC: %X:%X:%X:%X:%X:%X add !!\n",
						*mac_addr,*(mac_addr+1),*(mac_addr+2),*(mac_addr+3),*(mac_addr+4),*(mac_addr+5) );
				#endif
			}
			else{
				/* MAC pool is full */
				#ifdef CONFIG_BRIDGE_IGMPP_PROCFS_DEBUG
				uint8_t *ip8_addr;
				trans_32to8(&ip32_addr, &ip8_addr);
				printk(KERN_INFO "[BR_IGMPP_PROC]-> the pool of Group IP: %u.%u.%u.%u has been full !!\n",
													*ip8_addr, *(ip8_addr+1), *(ip8_addr+2), *(ip8_addr+3));
				#endif
			}
		}			
	}else{
		/* check group pool */
		int groupPoolIdx;
		groupPoolIdx = check_GROUP_pool(pt);
		if(groupPoolIdx >= 0){
			/* add group and host */
			/* add group */
			add_GROUP(pt, groupPoolIdx, ip32_addr);
			#ifdef CONFIG_BRIDGE_IGMPP_PROCFS_DEBUG
			uint8_t *ip8_addr;
			trans_32to8(&ip32_addr, &ip8_addr);
			printk(KERN_INFO "[BR_IGMPP_PROC]-> Group IP: %u.%u.%u.%u add !!\n",
												*ip8_addr, *(ip8_addr+1), *(ip8_addr+2), *(ip8_addr+3));
			#endif
			/* add host */
			add_MAC_2_pool(pt, groupPoolIdx, 0, mac_addr);
			#ifdef CONFIG_BRIDGE_IGMPP_PROCFS_DEBUG
			printk(KERN_INFO "[BR_IGMPP_PROC]-> MAC: %X:%X:%X:%X:%X:%X add !!\n",
						*mac_addr,*(mac_addr+1),*(mac_addr+2),*(mac_addr+3),*(mac_addr+4),*(mac_addr+5) );
			#endif

		}else {
			/* group pool is full */
			#ifdef CONFIG_BRIDGE_IGMPP_PROCFS_DEBUG
			printk(KERN_INFO "[BR_IGMPP_PROC]-> the pool of group on \" %s \" has been full !!", p->dev->name);
			#endif
		}
	}

	return;
}
static void table_remove(struct port_igmpp_table_t *pt, 
						uint32_t ip32_addr, 
						unsigned char * mac_addr)
{
	/* search group IP */
	int groupIdx,ipIdx;
	groupIdx = search_group_IP(pt, ip32_addr);
	if(groupIdx >= 0){
		/* search list MAC */
		ipIdx = search_list_MAC(pt, groupIdx, mac_addr);
		if(ipIdx >= 0){
			/* remove MAC and check group member*/
			remove_MAC_from_pool(pt, groupIdx, ipIdx);
			#ifdef CONFIG_BRIDGE_IGMPP_PROCFS_DEBUG
			printk(KERN_INFO "[BR_IGMPP_PROC]-> MAC: %X:%X:%X:%X:%X:%X remove !!\n",
						*mac_addr,*(mac_addr+1),*(mac_addr+2),*(mac_addr+3),*(mac_addr+4),*(mac_addr+5) );
			#endif
			check_GROUP_is_empty_and_remove(pt, groupIdx);
		}else{
			#ifdef CONFIG_BRIDGE_IGMPP_PROCFS_DEBUG
			printk(KERN_INFO "[BR_IGMPP_PROC]-> MAC: %X:%X:%X:%X:%X:%X does't exist !!\n",
						*mac_addr,*(mac_addr+1),*(mac_addr+2),*(mac_addr+3),*(mac_addr+4),*(mac_addr+5) );
			#endif
		}
	}else{
		#ifdef CONFIG_BRIDGE_IGMPP_PROCFS_DEBUG
		uint8_t *ip8_addr;
		trans_32to8(&ip32_addr, &ip8_addr);
		printk(KERN_INFO "[BR_IGMPP_PROC]-> Group IP: %u.%u.%u.%u does't exist !!\n",
											*ip8_addr, *(ip8_addr+1), *(ip8_addr+2), *(ip8_addr+3));
		#endif
	}

	return;
}
/* For different platform, 
 * kernel API may don't support strcmp() originally, 
 * we copy strcmp() and named strcmp1() for conveniently.
 *
 * strcmp - Compare two strings
 * @cs: One string
 * @ct: Another string
 */
static int strcmp1(char * cs,char * ct)
{
	register signed char __res;
	while (1) {
		if ((__res = *cs - *ct++) != 0 || !*cs++)
			break;
	}
	return __res;
}

/* For different platform, 
 * kernel API may don't  support strspn() originally, 
 * we copy strspn() and named strspn1() for conveniently.
 *
 * strspn1 - Calculate the length of the initial substring of @s which only
 *      contain letters in @accept
 * @s: The string to be searched
 * @accept: The string to search for
 */
static size_t strspn1(char *s, char *accept)
{
	char *p;
	char *a;
	size_t count = 0;

	for (p = s; *p != '\0'; ++p) {
		for (a = accept; *a != '\0'; ++a) {
			if (*p == *a)
				break;
		}
		if (*a == '\0')
			return count;
		++count;
	}
	return count;
}
/* check_str()
 * type => 0, only accept ".0123456789"
 * type => 1, only accept ":0123456789ABCDEFabcdef"
 */
static int check_str(char *s, int type)
{
	char * accept = NULL;
	switch (type){
		case 0:
			accept = IP_ACCEPT_CHAR;	
			break;
		case 1:
			accept = MAC_ACCEPT_CHAR;
			break;
	}

	if (accept != NULL)		
		if ( strlen(s) == strspn1(s, accept) )
			return 0;
	return (-1);
}

static struct net_bridge_port * search_device(struct net_bridge * br, char* name)
{
	struct net_bridge_port *p;

	p = br->port_list;
	while(p != NULL){

		if (strcmp1(p->dev->name, name) == 0 ){
			return p;
		}
		p = p->next;
	}
	return NULL;
}
static int get_MAC(unsigned char * mac_addr, struct net_bridge *br, uint32_t ip_addr)
{
	struct br_mac_table_t *tlist;
	int find = -1;
	read_lock(&br->br_mac_table_rwlock);
	list_for_each_entry(tlist,&(br->br_mac_table.list), list){
		if ( tlist->ip_addr == ip_addr){
			find = 0;
			int i;
			for (i =0; i<6; i++)
				mac_addr[i] = tlist->mac_addr[i];
			break;
		}
	}
	read_unlock(&br->br_mac_table_rwlock);
	return find;
}
/* enable each port's port_igmpp_table */
static void table_all_enable(struct net_bridge *br)
{
	/* set bridge */
	atomic_set(&br->br_igmpp_table_enable, 1);
	/* set each port */
	struct net_bridge_port *p;
	p = br->port_list;
	while( p != NULL){
		write_lock(&p->port_igmpp_table_rwlock);
		memset( &p->port_igmpp_table, 0, sizeof( struct port_igmpp_table_t));
		p->port_igmpp_table.enable = 1;
		write_unlock(&p->port_igmpp_table_rwlock);
		p=p->next;
	}
}
/* disable each port's port_igmpp_table */
static void table_all_disable(struct net_bridge *br)
{
	/* set bridge */
	atomic_set(&br->br_igmpp_table_enable, 0);
	/* set each port */
	struct net_bridge_port *p;
	p = br->port_list;
	while( p != NULL){
		write_lock(&p->port_igmpp_table_rwlock);
		p->port_igmpp_table.enable = 0;
		write_unlock(&p->port_igmpp_table_rwlock);
		p=p->next;
	}
}
static int table_setwl(struct net_bridge *br, char * name)
{
	struct net_bridge_port *hit_port;
	hit_port = search_device(br, name);
	if (hit_port != NULL){
		atomic_set(&hit_port->wireless_interface, 1);
		return 0;
	}else 
		return (-1);
}
static int table_unsetwl(struct net_bridge *br, char * name)
{
	struct net_bridge_port *hit_port;
	hit_port = search_device(br, name);
	if (hit_port != NULL){
		atomic_set(&hit_port->wireless_interface, 0);
		return 0;
	}else 
		return (-1);
}
static int proc_write_br_igmpp (struct file *file, const char *buf,
		unsigned long count, void *data)
{
	int len;
	if(count > MESSAGE_LENGTH) {len = MESSAGE_LENGTH;}
	else {len = count; }
	char message[len];
	if(copy_from_user(message, buf, len))
		return -EFAULT;
	message[len-1] = '\0';

	/* split input message that get from user space 
	 * token[0] => action token --> add or remove 
	 * token[1] => multicast group IP address 
	 * token[2] => MAC address of host IP  
	 */
	char *msgDelim = MESSAGE_DELIM;
	char *pmesg = message ;
	char *action_token, *action;

	action_token = strsep(&pmesg, msgDelim);

	struct net_bridge *br = (struct net_bridge *) data;

	/* ============================  enable each port's port_igmpp_table ====================*/
	/* NOTE: Each port_igmpp_table will be clean.								 			 */
	action = ACTION_ENABLE_TABLE;
	if ( memcmp(action_token, action, sizeof(ACTION_ENABLE_TABLE) )== 0){
		table_all_enable(br);
		#ifdef CONFIG_BRIDGE_IGMPP_PROCFS_DEBUG
		printk(KERN_INFO "[BR_IGMPP_PROC]->ENABLE %s\n",br->dev.name);
		#endif
	}

	/* ============================  disable each port's port_igmpp_table ===================*/
	action = ACTION_DISABLE_TABLE;
	if ( memcmp(action_token, action, sizeof(ACTION_DISABLE_TABLE) )== 0){
		table_all_disable(br);
		#ifdef CONFIG_BRIDGE_IGMPP_PROCFS_DEBUG
		printk(KERN_INFO "[BR_IGMPP_PROC]->DISABLE %s\n",br->dev.name);
		#endif 
	}

	/* ============================  set wireless ============================*/
	action = ACTION_SET_WL;
	if ( memcmp(action_token, action, sizeof(ACTION_SET_WL) )== 0){
		if (table_setwl(br,pmesg) == 0){
			#ifdef CONFIG_BRIDGE_IGMPP_PROCFS_DEBUG
			printk(KERN_INFO "[BR_IGMPP_PROC]->SETWL-> %s \n",pmesg);
			#endif
		}else{
			printk(KERN_INFO "[BR_IGMPP_PROC]->WARNING SETWL FAILURE-> %s\n",pmesg);
			goto proc_write_br_igmpp_out;
		}
	}
	
	/* ============================  unset wireless ==========================*/
	action = ACTION_UNSET_WL;
	if ( memcmp(action_token, action, sizeof(ACTION_UNSET_WL) )== 0){
		if (table_unsetwl(br,pmesg) == 0){
			#ifdef CONFIG_BRIDGE_IGMPP_PROCFS_DEBUG
			printk(KERN_INFO "[BR_IGMPP_PROC]->UNSETWL-> %s \n",pmesg);
			#endif
		}else{
			printk(KERN_INFO "[BR_IGMPP_PROC]->WARNING UNSETWL FAILURE-> %s\n",pmesg);
			goto proc_write_br_igmpp_out;
		}
	}

	/* ============================  add =====================================*/
	action = ACTION_ADD;
	if ( memcmp(action_token, action, sizeof(ACTION_ADD) )== 0){
		/********** START: processing input string **********/
		char *token[2];
		int i;
		for(i=0; i<2; ++i) 
			token[i] = strsep(&pmesg, msgDelim);
		/* split multicast group IP address */
		uint32_t grp_ip32_addr;
		if (check_str(token[0], 0) == 0 )
			split_IP(&grp_ip32_addr,token[0]);
		else{
			#ifdef CONFIG_BRIDGE_IGMPP_PROCFS_DEBUG
			printk(KERN_INFO "[BR_IGMPP_PROC]-> Group IP address:%s is illegal !! \n",token[0]);
			#endif
			goto proc_write_br_igmpp_out;
		}

		/* Judge address of Host is IP adress or MAC address */
		unsigned char mac_addr[6];
		uint32_t host_ip32_addr;
		read_lock(&br->br_mac_table_rwlock);
		if (atomic_read(&br->br_mac_table_enable) == 1 ){
			/* split host IP address */
			read_unlock(&br->br_mac_table_rwlock);
			if (check_str(token[1], 0) == 0 )
				split_IP(&host_ip32_addr,token[1]);
			else{
				#ifdef CONFIG_BRIDGE_IGMPP_PROCFS_DEBUG
				printk(KERN_INFO "[BR_IGMPP_PROC]-> Host IP address: %s is illegal !!\n",token[1]);
				printk(KERN_INFO "br_mac_table_enable is 1, Host address only accept char: \".1234567890\"\n");
				#endif
				goto proc_write_br_igmpp_out;
			}
			/* look for MAC address with IP */
			if ( get_MAC(mac_addr, br, host_ip32_addr) != 0 ){
				uint8_t *ip8_addr;
				trans_32to8(&host_ip32_addr, &ip8_addr);
                printk(KERN_INFO "[BR_IGMPP_PROC]-> Could not find MAC with IP: %u.%u.%u.%u !!\n",
									*ip8_addr, *(ip8_addr+1), *(ip8_addr+2), *(ip8_addr+3));
				goto proc_write_br_igmpp_out;
			}
		}else {
			/* split host MAC address */
			read_unlock(&br->br_mac_table_rwlock);
			if (check_str(token[1], 1) == 0 )
				split_MAC(mac_addr, token[1]);
			else{
				#ifdef CONFIG_BRIDGE_IGMPP_PROCFS_DEBUG
				printk(KERN_INFO "[BR_IGMPP_PROC]-> Host MAC address: %s is illegal !!\n",token[1]);
				printk(KERN_INFO "br_mac_table_enable is 0, Host address only accept char: \":1234567890ABCDEFabcdef\"\n");
				#endif
				goto proc_write_br_igmpp_out;
			}
		}
		/********** END: processing input string **********/

		/* searching bridge_fdb_entry */
		struct net_bridge_fdb_entry *hit_fdb_entry;
		hit_fdb_entry = br_fdb_get(br, mac_addr);
		if (hit_fdb_entry != NULL){
			if(atomic_read(&hit_fdb_entry->dst->wireless_interface) == 1){
				#ifdef CONFIG_BRIDGE_IGMPP_PROCFS_DEBUG
				printk(KERN_INFO "[BR_IGMPP_PROC]-> ADD ...\n");
				#endif
				write_lock(&hit_fdb_entry->dst->port_igmpp_table_rwlock);
				table_add(hit_fdb_entry->dst, &hit_fdb_entry->dst->port_igmpp_table, 
							grp_ip32_addr, mac_addr);
				write_unlock(&hit_fdb_entry->dst->port_igmpp_table_rwlock);
				#ifdef CONFIG_BRIDGE_IGMPP_PROCFS_DEBUG
				printk(KERN_INFO "[BR_IGMPP_PROC]-> ADD DONE !!\n");
				/* make sure there's only one record existed */
				printk(KERN_INFO "[BR_IGMPP_PROC]-> checking and clean all other port_igmpp_table ... \n");
				#endif
				struct net_bridge_port *p;
				p = br->port_list;
				while( p != NULL){
					if(hit_fdb_entry->dst->port_no != p->port_no){
						write_lock(&p->port_igmpp_table_rwlock);
						table_remove(&p->port_igmpp_table, grp_ip32_addr, mac_addr);
						write_unlock(&p->port_igmpp_table_rwlock);
					}
					p=p->next;
				}
				#ifdef CONFIG_BRIDGE_IGMPP_PROCFS_DEBUG
				printk(KERN_INFO "[BR_IGMPP_PROC]-> clean other port_igmpp_table DONE !!\n");
				#endif
			}else{
				#ifdef CONFIG_BRIDGE_IGMPP_PROCFS_DEBUG
				uint8_t *ip8_addr;
				trans_32to8(&host_ip32_addr, &ip8_addr);
				printk(KERN_INFO "[BR_IGMPP_PROC]-> The host IP: %u.%u.%u.%u is belong to wired interface(port) !!\n",
									*ip8_addr, *(ip8_addr+1), *(ip8_addr+2), *(ip8_addr+3));
				#endif
				br_fdb_put(hit_fdb_entry);
				goto proc_write_br_igmpp_out;
			}
			br_fdb_put(hit_fdb_entry);
		}else{
			printk(KERN_INFO "The return value of br_fdb_get() is NULL -> MAC: %X:%X:%X:%X:%X:%X \n",
					mac_addr[0], mac_addr[1], mac_addr[2], mac_addr[3], mac_addr[4], mac_addr[5] );
			goto proc_write_br_igmpp_out;
		}
	}

	/* ============================  remove  =================================*/
	action = ACTION_REMOVE;
	if ( memcmp(action_token, action, sizeof(ACTION_REMOVE) ) == 0){
		/********** START: processing input string **********/
		char *token[2];
		int i;
		for(i=0; i<2; ++i) 
			token[i] = strsep(&pmesg, msgDelim);
		/* split multicast group IP address */
		uint32_t grp_ip32_addr;
		if (check_str(token[0], 0) == 0 )
			split_IP(&grp_ip32_addr,token[0]);
		else{
			#ifdef CONFIG_BRIDGE_IGMPP_PROCFS_DEBUG
			printk(KERN_INFO "[BR_IGMPP_PROC]-> Group IP address:%s is illegal !! \n",token[0]);
			#endif
			goto proc_write_br_igmpp_out;
		}
	
		/* Judge address of Host is IP adress or MAC address */
		unsigned char mac_addr[6];
		uint32_t host_ip32_addr;
		read_lock(&br->br_mac_table_rwlock);
		if (atomic_read(&br->br_mac_table_enable) == 1 ){
			/* split host IP address */
			read_unlock(&br->br_mac_table_rwlock);

			if (check_str(token[1], 0) == 0 )
				split_IP(&host_ip32_addr,token[1]);
			else{
				#ifdef CONFIG_BRIDGE_IGMPP_PROCFS_DEBUG
				printk(KERN_INFO "[BR_IGMPP_PROC]-> Host IP address: %s is illegal !!\n",token[1]);
				printk(KERN_INFO "br_mac_table_enable is 1, Host address only accept char: \".1234567890\"\n");
				#endif
				goto proc_write_br_igmpp_out;
			}
			/* look for MAC address with IP */
			if ( get_MAC(mac_addr, br, host_ip32_addr) != 0 ){
				uint8_t *ip8_addr;
				trans_32to8(&host_ip32_addr, &ip8_addr);
                printk(KERN_INFO "[BR_IGMPP_PROC]-> Could not find MAC with IP: %u.%u.%u.%u !!\n",
									*ip8_addr, *(ip8_addr+1), *(ip8_addr+2), *(ip8_addr+3));
				goto proc_write_br_igmpp_out;
			}
		}else {
			/* split host MAC address */
			read_unlock(&br->br_mac_table_rwlock);
			if (check_str(token[1], 1) == 0 )
				split_MAC(mac_addr, token[1]);
			else{
				#ifdef CONFIG_BRIDGE_IGMPP_PROCFS_DEBUG
				printk(KERN_INFO "[BR_IGMPP_PROC]-> Host MAC address: %s is illegal !!\n",token[1]);
				printk(KERN_INFO "br_mac_table_enable is 0, Host address only accept char: \":1234567890ABCDEFabcdef\"\n");
				#endif
				goto proc_write_br_igmpp_out;
			}
		}
		/********** END: processing input string **********/

		/* searching bridge_fdb_entry */
		struct net_bridge_fdb_entry *hit_fdb_entry;
		hit_fdb_entry = br_fdb_get(br, mac_addr);
		if (hit_fdb_entry != NULL){
			if(atomic_read(&hit_fdb_entry->dst->wireless_interface) == 1){
				#ifdef CONFIG_BRIDGE_IGMPP_PROCFS_DEBUG
				printk(KERN_INFO "[BR_IGMPP_PROC]-> REMOVE ...\n");
				#endif
				write_lock(&hit_fdb_entry->dst->port_igmpp_table_rwlock);
				table_remove(&hit_fdb_entry->dst->port_igmpp_table, grp_ip32_addr, mac_addr);
				write_unlock(&hit_fdb_entry->dst->port_igmpp_table_rwlock);
				#ifdef CONFIG_BRIDGE_IGMPP_PROCFS_DEBUG
				printk(KERN_INFO "[BR_IGMPP_PROC]-> REMOVE DONE !!\n");
				#endif
			}else{
				#ifdef CONFIG_BRIDGE_IGMPP_PROCFS_DEBUG
				uint8_t *ip8_addr;
				trans_32to8(&host_ip32_addr, &ip8_addr);
				printk(KERN_INFO "[BR_IGMPP_PROC]-> The host IP: %u.%u.%u.%u is belong to wired interface(port) !!\n",
									*ip8_addr, *(ip8_addr+1), *(ip8_addr+2), *(ip8_addr+3));
				#endif
				br_fdb_put(hit_fdb_entry);
				goto proc_write_br_igmpp_out;
			}
			br_fdb_put(hit_fdb_entry);
		}else{
			printk(KERN_INFO "The return value of br_fdb_get() is NULL -> MAC: %X:%X:%X:%X:%X:%X \n",
					mac_addr[0], mac_addr[1], mac_addr[2], mac_addr[3], mac_addr[4], mac_addr[5] );
			goto proc_write_br_igmpp_out;
		}
	}

	/* ============================= END ======================================*/
proc_write_br_igmpp_out:
	return len;

}


/* procfs write/read for br_mac_table */

static int proc_read_br_mac (char *buf, char **start, off_t offset,
		int len, int *eof, void *data)
{

	int count =0;
	struct net_bridge *br = (struct net_bridge *) data;
	struct br_mac_table_t *tlist;

	read_lock(&br->br_mac_table_rwlock); //reader spin lock

	printk( "*********************************\n");
	printk( "* br->br_mac_table_enable : %d\n",atomic_read(&br->br_mac_table_enable));
	printk( "*********************************\n");
	printk( "IP address         MAC           \n");

	list_for_each_entry(tlist, &(br->br_mac_table.list), list){
		uint32_t ip32_addr = tlist->ip_addr;
		uint8_t *ip8_addr;
		trans_32to8(&ip32_addr, &ip8_addr);
		printk("%u.%u.%u.%u \t %X:%X:%X:%X:%X:%X \n", 
			*ip8_addr, *(ip8_addr+1), *(ip8_addr+2), *(ip8_addr+3),
			tlist->mac_addr[0],	tlist->mac_addr[1], tlist->mac_addr[2], tlist->mac_addr[3],	
			tlist->mac_addr[4], tlist->mac_addr[5]
		);
	}

	read_unlock(&br->br_mac_table_rwlock); // reader spin unlock

	*eof = 1;
	return count;
}

static int proc_write_br_mac (struct file *file, const char *buf,
		unsigned long count, void *data)
{

	int len;
	if(count > MESSAGE_LENGTH) {len = MESSAGE_LENGTH;}
	else {len = count; }
	char message[len];
	if(copy_from_user(message, buf, len))
		return -EFAULT;
	message[len-1] = '\0';

	char *action;
	struct net_bridge *br = (struct net_bridge *) data;
	write_lock(&br->br_mac_table_rwlock); // writer spin lock

	/* ============================  enable br_mac_table ===================*/
	/* NOTE: The br_mac_table will be clean.								  */
	action = ACTION_ENABLE_TABLE;
	if ( memcmp(message, action, sizeof(ACTION_ENABLE_TABLE) )== 0){
		/* reset br_mac_table */
		struct br_mac_table_t *tlist;
		struct list_head *pos, *q;
		list_for_each_safe(pos, q, &(br->br_mac_table.list) ){
			tlist= list_entry(pos, struct br_mac_table_t, list);
			list_del(pos);
			kfree(tlist);
		}
		memset( &(br->br_mac_table), 0, sizeof( struct br_mac_table_t));
		INIT_LIST_HEAD(&br->br_mac_table.list);
		atomic_set(&br->br_mac_table_enable, 1);
		#ifdef CONFIG_BRIDGE_IGMPP_PROCFS_DEBUG
		printk(KERN_INFO "[BR_MAC_PROC]->ENABLE %s\n", br->dev.name);
		printk(KERN_INFO "br_mac_table_enable = %d\n", atomic_read(&br->br_mac_table_enable) );
		#endif
	}

	/* ============================  disable br_mac_table ===================*/
	action = ACTION_DISABLE_TABLE;
	if ( memcmp(message, action, sizeof(ACTION_DISABLE_TABLE) )== 0){
		atomic_set(&br->br_mac_table_enable, 0);
		#ifdef CONFIG_BRIDGE_IGMPP_PROCFS_DEBUG
		printk(KERN_INFO "[BR_MAC_PROC]->DISABLE %s\n", br->dev.name);
		printk(KERN_INFO "br_mac_table_enable = %d\n", atomic_read(&br->br_mac_table_enable) );
		#endif
	}

	write_unlock(&br->br_mac_table_rwlock); // writer spin unlock
	return len;
}

#endif

#ifdef CONFIG_BRIDGE_FORWARD_CTRL
static int proc_read_br_forward(char * buf, char ** start, off_t offset,
		int len, int * eof, void * data)
{
	struct net_bridge * br = (struct net_bridge *)data;
	char * p = buf;
	p += sprintf(p, "%d\n", atomic_read(&br->br_forward));
	*eof = 1;
	return p - buf;
}

static int proc_write_br_forward(struct file * file, const char * buf,
		unsigned long count, void * data)
{
	struct net_bridge * br = (struct net_bridge *)data;
	if (count < 1) return -EINVAL;
	atomic_set(&br->br_forward, buf[0] == '0' ? 0 : 1);
	return count;
}
#endif

int br_add_bridge(char *name)
{
	struct net_bridge *br;
	struct net_device *dev;
	int err;
#ifdef CONFIG_BRIDGE_FORWARD_CTRL
	char proc_name[32];
#endif

	if ((br = new_nb(name)) == NULL)
		return -ENOMEM;

	dev = &br->dev;
	if (strchr(dev->name, '%')) {
		err = dev_alloc_name(dev, dev->name);
		if (err < 0)
			goto  out;
	}

	err = register_netdevice(dev);
	if (err) goto out;
	br_inc_use_count();

#ifdef CONFIG_BRIDGE_FORWARD_CTRL
	atomic_set(&br->br_forward, 1);
	snprintf(proc_name, sizeof(proc_name), "net/br_forward_%s", name);
	br->br_proc = create_proc_entry(proc_name, 0, 0);
	if (br->br_proc == NULL) return -ENOMEM;
	br->br_proc->data = (void *)br;
	br->br_proc->read_proc = proc_read_br_forward;
	br->br_proc->write_proc = proc_write_br_forward;
#endif
#ifdef CONFIG_BRIDGE_IGMP_SNOOP
	char igmp_ap_proc_name[32];
	snprintf(igmp_ap_proc_name, sizeof(igmp_ap_proc_name), "net/br_igmp_ap_%s", name);
	br->br_igmp_ap_proc = create_proc_entry (igmp_ap_proc_name, 0, 0);
	if (br->br_igmp_ap_proc == NULL)
		return -ENOMEM;
	br->br_igmp_ap_proc->data = (void*)br;
	br->br_igmp_ap_proc->read_proc = proc_read_br_igmp_ap;
	br->br_igmp_ap_proc->write_proc = proc_write_br_igmp_ap;	
    INIT_LIST_HEAD(&br->br_mac_ap_table.list);


	snprintf(igmp_ap_proc_name, sizeof(igmp_ap_proc_name), "net/br_query_%s", name);
	br->br_query_proc = create_proc_entry (igmp_ap_proc_name, 0, 0);
	if (br->br_query_proc == NULL)
		return -ENOMEM;
	br->br_query_proc->data = (void*)br;
	br->br_query_proc->read_proc = proc_read_br_query;
	br->br_query_proc->write_proc = proc_write_br_query;

	
#endif //CONFIG_BRIDGE_IGMP_SNOOP

#ifdef CONFIG_BRIDGE_IGMPP_PROCFS
	char igmpp_proc_name[32];
	snprintf(igmpp_proc_name, sizeof(igmpp_proc_name), "net/br_igmpp_%s", name);
	br->br_igmpp_proc = create_proc_entry (igmpp_proc_name, 0, 0);
	if (br->br_igmpp_proc == NULL)
		return -ENOMEM;
	br->br_igmpp_proc->data = (void*)br;
	br->br_igmpp_proc->read_proc = proc_read_br_igmpp;
	br->br_igmpp_proc->write_proc = proc_write_br_igmpp;

	snprintf(igmpp_proc_name, sizeof(igmpp_proc_name), "net/br_mac_%s", name);
	br->br_mac_proc = create_proc_entry (igmpp_proc_name, 0, 0);
	if (br->br_mac_proc == NULL)
		return -ENOMEM;
	br->br_mac_proc->data = (void*)br;
	br->br_mac_proc->read_proc = proc_read_br_mac;
	br->br_mac_proc->write_proc = proc_write_br_mac;

    INIT_LIST_HEAD(&br->br_mac_table.list);
#endif

 out:
	return err;
}

int br_del_bridge(char *name)
{
	struct net_device *dev;
	struct net_bridge *br;
#ifdef CONFIG_BRIDGE_FORWARD_CTRL
	char proc_name[32];
#endif

	dev = __dev_get_by_name(name);
	if (!dev)
		return -ENXIO;

	if (dev->hard_start_xmit != br_dev_xmit)
		return -EPERM;

	if (dev->flags & IFF_UP)
		return -EBUSY;

	br = dev->priv;
	BUG_ON(&br->dev != dev);

#ifdef CONFIG_BRIDGE_FORWARD_CTRL
	snprintf(proc_name, sizeof(proc_name), "net/br_forward_%s", name);
	remove_proc_entry(proc_name, 0);
#endif
#ifdef CONFIG_BRIDGE_IGMPP_PROCFS
	char igmpp_proc_name[32];
	snprintf(igmpp_proc_name, sizeof(igmpp_proc_name), "net/br_igmpp_%s", name);
	remove_proc_entry(igmpp_proc_name, 0);

	snprintf(igmpp_proc_name, sizeof(igmpp_proc_name), "net/br_mac_%s", name);
	remove_proc_entry(igmpp_proc_name, 0);
#endif

#ifdef CONFIG_BRIDGE_IGMP_SNOOP
	char igmp_ap_proc_name[32];
	snprintf(igmp_ap_proc_name, sizeof(igmp_ap_proc_name), "net/br_igmp_ap_%s", name);
	remove_proc_entry(igmp_ap_proc_name, 0);

	snprintf(igmp_ap_proc_name, sizeof(igmp_ap_proc_name), "net/br_ap_mac_%s", name);
	remove_proc_entry(igmp_ap_proc_name, 0);
	
	snprintf(igmp_ap_proc_name, sizeof(igmp_ap_proc_name), "net/net/br_query_%s", name);
	remove_proc_entry(igmp_ap_proc_name, 0);
#endif //CONFIG_BRIDGE_IGMP_SNOOP
	del_ifs(br);

	unregister_netdevice(dev);

	kfree(br);
	br_dec_use_count();

	return 0;
}

int br_add_if(struct net_bridge *br, struct net_device *dev)
{
	struct net_bridge_port *p;

	if (dev->br_port != NULL)
		return -EBUSY;

	if (dev->flags & IFF_LOOPBACK || dev->type != ARPHRD_ETHER)
		return -EINVAL;

	if (dev->hard_start_xmit == br_dev_xmit)
		return -ELOOP;

	if (!is_valid_ether_addr(dev->dev_addr))
		return -EADDRNOTAVAIL;

	dev_hold(dev);
	write_lock_bh(&br->lock);
	if ((p = new_nbp(br, dev)) == NULL) {
		write_unlock_bh(&br->lock);
		dev_put(dev);
		return -EXFULL;
	}

	dev_set_promiscuity(dev, 1);

	br_stp_recalculate_bridge_id(br);
#ifdef CONFIG_BRIDGE_MAT
	{
	unsigned long ipaddr = inet_select_addr(dev, 0 , 0);
	br_fdb_insert(br, p, dev->dev_addr, ipaddr, 1);
	}
#else
	br_fdb_insert(br, p, dev->dev_addr, 1);
#endif /*CONFIG_BRIDGE_MAT*/
	if ((br->dev.flags & IFF_UP) && (dev->flags & IFF_UP))
		br_stp_enable_port(p);
	write_unlock_bh(&br->lock);

	return 0;
}

int br_del_if(struct net_bridge *br, struct net_device *dev)
{
	int retval;

	br_write_lock_bh(BR_NETPROTO_LOCK);
	write_lock(&br->lock);
	retval = __br_del_if(br, dev);
	br_stp_recalculate_bridge_id(br);
	write_unlock(&br->lock);
	br_write_unlock_bh(BR_NETPROTO_LOCK);

	return retval;
}

int br_get_bridge_ifindices(int *indices, int num)
{
	struct net_device *dev;
	int i = 0;

	for (dev = dev_base; i < num && dev != NULL; dev = dev->next) {
		if (dev->hard_start_xmit == br_dev_xmit)
			indices[i++] = dev->ifindex;
	}

	return i;
}

/* called under ioctl_lock */
void br_get_port_ifindices(struct net_bridge *br, int *ifindices)
{
	struct net_bridge_port *p;

	p = br->port_list;
	while (p != NULL) {
		ifindices[p->port_no] = p->dev->ifindex;
		p = p->next;
	}
}
