/*
 *	Linux ethernet bridge
 *
 *	Authors:
 *	Lennert Buytenhek		<buytenh@gnu.org>
 *
 *	$Id: br_private.h,v 1.1.1.1 2006-10-13 02:42:51 steven Exp $
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License
 *	as published by the Free Software Foundation; either version
 *	2 of the License, or (at your option) any later version.
 */

#ifndef _BR_PRIVATE_H
#define _BR_PRIVATE_H

#include <linux/netdevice.h>
#include <linux/miscdevice.h>
#include <linux/if_bridge.h>

#ifdef CONFIG_BRIDGE_IGMPP_PROCFS
#include <linux/list.h>
#endif

#include "br_private_timer.h"

#ifdef CONFIG_BRIDGE_MAT
#include <linux/ip.h>
#include <linux/if_arp.h>
#include <linux/if_ether.h>
#endif /*CONFIG_BRIDGE_MAT*/

#ifdef CONFIG_BRIDGE_MAT
#define WLAN         "ra0"
#define ETHERNET     "eth0"
#define MAT_DBG      0

#define AP           0
#define APC          1

struct postarphdr {
	unsigned char		ar_sha[ETH_ALEN];	/* sender hardware address	*/
	unsigned char		ar_sip[4];		/* sender IP address		*/
	unsigned char		ar_tha[ETH_ALEN];	/* target hardware address	*/
	unsigned char		ar_tip[4];		/* target IP address		*/
};
#endif /*CONFIG_BRIDGE_MAT*/

#define BR_HASH_BITS 8
#define BR_HASH_SIZE (1 << BR_HASH_BITS)

#define BR_HOLD_TIME (1*HZ)

typedef struct bridge_id bridge_id;
typedef struct mac_addr mac_addr;
typedef __u16 port_id;

#ifdef CONFIG_BRIDGE_IGMPP_PROCFS
#define BR_IGMPP_MSG_ERROR		0
#define BR_IGMPP_MSG_OK			1
#define BR_IGMPP_MSG_INFO		2	
#define BR_IGMPP_MSG_WARNING	3	

#define MESSAGE_LENGTH			80
#define MESSAGE_DELIM			" \t"
#define MESSAGE_ARGC			3
#define IP_DELIM				"."
#define IP_ACCEPT_CHAR			".0123456789"
#define MAC_DELIM				":"
#define MAC_ACCEPT_CHAR			":0123456789ABCDEFabcdef"
#define ACTION_ADD				"add"
#define ACTION_REMOVE			"remove"
#define ACTION_SET_WL			"setwl"
#define ACTION_UNSET_WL			"unsetwl"
#define ACTION_ENABLE_TABLE		"enable"
#define ACTION_DISABLE_TABLE	"disable"

#define HOSTLIST_NUMBER			8		
#define GROUPLIST_NUMBER		8						

struct port_igmpp_mac_t {
	int				used;
	unsigned char	mac_addr[6];			
};

struct port_igmpp_group_t {
	int				used;
	uint32_t		ip_addr;
	struct	port_igmpp_mac_t host_list[HOSTLIST_NUMBER];
};

struct port_igmpp_table_t {
	int				enable;	
	struct port_igmpp_group_t group_list[GROUPLIST_NUMBER];
};

struct br_mac_table_t {
	struct list_head list;
	uint32_t		ip_addr;
	unsigned char	mac_addr[6];
};

#endif
#ifdef CONFIG_BRIDGE_IGMP_SNOOP
#define MESSAGE_LENGTH			80
#define MESSAGE_DELIM			" \t"
#define MESSAGE_ARGC			3
#define ACTION_SET_WL			"setwl"
#define ACTION_UNSET_WL			"unsetwl"
#define ACTION_ENABLE_TABLE		"enable"
#define ACTION_DISABLE_TABLE	"disable"
#define HOSTLIST_NUMBER			8		
#define GROUPLIST_NUMBER		8						

struct port_igmp_ap_mac_t { //sta'mac
	int				used;
	unsigned char	mac_addr[6];			
};

struct port_igmp_ap_group_t {
	int				used;		
	uint32_t		ip_addr;
	uint8_t        in_special_query;//if this port is in special_query state
	int	port_aging_time;
    uint8_t max_response_time;
	struct	port_igmp_ap_mac_t host_list[HOSTLIST_NUMBER];
};

struct port_igmp_ap_table_t {	
	int				enable;	
	struct port_igmp_ap_group_t group_list[30];
};

struct br_mac_ap_table_t {
	struct list_head list;
	uint32_t		ip_addr;
	unsigned char	mac_addr[6];
};

#endif //CONFIG_BRIDGE_IGMP_SNOOP


struct bridge_id
{
	unsigned char	prio[2];
	unsigned char	addr[6];
};

struct mac_addr
{
	unsigned char	addr[6];
	unsigned char	pad[2];
};

struct net_bridge_fdb_entry
{
	struct net_bridge_fdb_entry	*next_hash;
	struct net_bridge_fdb_entry	**pprev_hash;
	atomic_t			use_count;
	mac_addr			addr;
	struct net_bridge_port		*dst;
	unsigned long			ageing_timer;
	unsigned			is_local:1;
	unsigned			is_static:1;
#ifdef CONFIG_BRIDGE_MAT
	unsigned int		ipaddr;
#endif /*CONFIG_BRIDGE_MAT*/
};

struct net_bridge_port
{
	struct net_bridge_port		*next;
	struct net_bridge		*br;
	struct net_device		*dev;
	int				port_no;

	/* STP */
	port_id				port_id;
	int				state;
	int				path_cost;
	bridge_id			designated_root;
	int				designated_cost;
	bridge_id			designated_bridge;
	port_id				designated_port;
	unsigned			topology_change_ack:1;
	unsigned			config_pending:1;
	int				priority;

	struct br_timer			forward_delay_timer;
	struct br_timer			hold_timer;
	struct br_timer			message_age_timer;

#ifdef CONFIG_BRIDGE_MULTICAST_BWCTRL
	unsigned int			bandwidth;
	unsigned int			accumulation;
	struct br_timer			bwctrl_timer;
#endif
#ifdef CONFIG_BRIDGE_PORT_FORWARD
	u8          port_forwarding;
#endif
#ifdef CONFIG_BRIDGE_IGMP_SNOOP	
	rwlock_t				port_igmp_ap_table_rwlock;
	struct port_igmp_ap_table_t port_igmp_ap_table;	// two-way array	
	atomic_t				ap_wireless_interface;
#endif	

#ifdef CONFIG_BRIDGE_IGMPP_PROCFS
	rwlock_t				port_igmpp_table_rwlock;
	struct port_igmpp_table_t port_igmpp_table;	// two-way array

	/* wireless_interface = 1, corresponding device of net_bridge_port is wireless device. 
	 * wireless_interface = 0, corresponding device of net_bridge_port is wired device.
	 * NOTE & TODO: This variable only change by IGMPProxy, should we modify br_add_if() 
	 * 				for initial this variable ?!
	 */
	atomic_t				wireless_interface;
#endif
#ifdef CONFIG_BRIDGE_REJECT_FORWARD_LIST
	u32			rejected_port_list;
#endif
};

struct net_bridge
{
	rwlock_t			lock;
	struct net_bridge_port		*port_list;
	struct net_device		dev;
	struct net_device_stats		statistics;
	rwlock_t			hash_lock;
	struct net_bridge_fdb_entry	*hash[BR_HASH_SIZE];
	struct timer_list		tick;

	/* STP */
	bridge_id			designated_root;
	int				root_path_cost;
	int				root_port;
	int				max_age;
	int				hello_time;
	int				forward_delay;
	bridge_id			bridge_id;
	int				bridge_max_age;
	int				bridge_hello_time;
	int				bridge_forward_delay;
	unsigned			stp_enabled:1;
	unsigned			topology_change:1;
	unsigned			topology_change_detected:1;

	struct br_timer			hello_timer;
	struct br_timer			tcn_timer;
	struct br_timer			topology_change_timer;
	struct br_timer			gc_timer;

	int				ageing_time;
	int				gc_interval;

#ifdef CONFIG_BRIDGE_FORWARD_CTRL
	atomic_t				br_forward;
	struct proc_dir_entry *	br_proc;
#endif
#ifdef CONFIG_BRIDGE_IGMP_SNOOP
	atomic_t					br_igmp_ap_table_enable; // for check each port_igmp_ap_table conveniently
	struct proc_dir_entry		*br_igmp_ap_proc; // port_igmpp_table I/O with user or processes

	rwlock_t					br_mac_table_ap_rwlock;
	struct br_mac_ap_table_t		br_mac_ap_table; // linking list structure	
	atomic_t					br_mac_ap_table_enable; // for check br_mac_table conveniently
	//struct proc_dir_entry		*br_mac_ap_proc; // br_mac_table I/O with user or processes

	struct proc_dir_entry		*br_query_proc;
	struct timer_list		igmptimer;
	atomic_t					br_igmp_AP_query_enable;
	int  igmp_enable;
	int  igmp_switch_query_timer;
#endif //CONFIG_BRIDGE_IGMP_SNOOP	
#ifdef CONFIG_BRIDGE_IGMPP_PROCFS
	atomic_t					br_igmpp_table_enable; // for check each port_igmpp_table conveniently
	struct proc_dir_entry		*br_igmpp_proc; // port_igmpp_table I/O with user or processes

	rwlock_t					br_mac_table_rwlock;
	struct br_mac_table_t		br_mac_table; // linking list structure	
	atomic_t					br_mac_table_enable; // for check br_mac_table conveniently
	struct proc_dir_entry		*br_mac_proc; // br_mac_table I/O with user or processes
#endif 
#ifdef CONFIG_BRIDGE_MAT
	unsigned char ap_mode;
#endif /*CONFIG_BRIDGE_MAT*/
};

extern struct notifier_block br_device_notifier;
extern unsigned char bridge_ula[6];

/* br.c */
extern void br_dec_use_count(void);
extern void br_inc_use_count(void);

/* br_device.c */
extern void br_dev_setup(struct net_device *dev);
extern int br_dev_xmit(struct sk_buff *skb, struct net_device *dev);
#ifdef CONFIG_BRIDGE_IGMP_SNOOP
extern int search_group_ap_IP(struct port_igmp_ap_table_t *pt, uint32_t ip32_addr);
#endif //CONFIG_BRIDGE_IGMP_SNOOP
/* br_fdb.c */
extern void br_fdb_changeaddr(struct net_bridge_port *p,
		       unsigned char *newaddr);
extern void br_fdb_cleanup(struct net_bridge *br);
extern void br_fdb_delete_by_port(struct net_bridge *br,
			   struct net_bridge_port *p);
#ifdef CONFIG_BRIDGE_MAT
extern struct net_bridge_fdb_entry *br_fdb_ip_get(struct net_bridge *br,
                    unsigned int addr);
#endif /*CONFIG_BRIDGE_MAT*/
extern struct net_bridge_fdb_entry *br_fdb_get(struct net_bridge *br,
					unsigned char *addr);
extern void br_fdb_put(struct net_bridge_fdb_entry *ent);
extern int  br_fdb_get_entries(struct net_bridge *br,
			unsigned char *_buf,
			int maxnum,
			int offset);
#ifdef CONFIG_BRIDGE_MAT
extern void br_fdb_insert(struct net_bridge *br,
		   struct net_bridge_port *source,
		   unsigned char *addr,
		   unsigned int ipaddr,
		   int is_local);
#else
extern void br_fdb_insert(struct net_bridge *br,
		   struct net_bridge_port *source,
		   unsigned char *addr,
		   int is_local);
#endif /*CONFIG_BRIDGE_MAT*/

/* br_forward.c */
extern void br_deliver(struct net_bridge_port *to,
		struct sk_buff *skb);
extern void br_forward(struct net_bridge_port *to,
		struct sk_buff *skb);
extern void br_flood_deliver(struct net_bridge *br,
		      struct sk_buff *skb,
		      int clone);
extern void br_flood_forward(struct net_bridge *br,
		      struct sk_buff *skb,
		      int clone);

/* br_if.c */
extern int br_add_bridge(char *name);
extern int br_del_bridge(char *name);
extern int br_add_if(struct net_bridge *br,
	      struct net_device *dev);
extern int br_del_if(struct net_bridge *br,
	      struct net_device *dev);
extern int br_get_bridge_ifindices(int *indices,
			    int num);
extern void br_get_port_ifindices(struct net_bridge *br,
			   int *ifindices);

#ifdef CONFIG_BRIDGE_IGMPP_PROCFS
extern int search_group_IP(struct port_igmpp_table_t *pt, uint32_t ip_addr);
#endif

/* br_input.c */
extern void br_handle_frame(struct sk_buff *skb);

/* br_ioctl.c */
extern int br_ioctl(struct net_bridge *br,
	     unsigned int cmd,
	     unsigned long arg0,
	     unsigned long arg1,
	     unsigned long arg2);
extern int br_ioctl_deviceless_stub(unsigned long arg);

/* br_stp.c */
extern int br_is_root_bridge(struct net_bridge *br);
extern struct net_bridge_port *br_get_port(struct net_bridge *br,
				    int port_no);
extern void br_init_port(struct net_bridge_port *p);
extern port_id br_make_port_id(struct net_bridge_port *p);
extern void br_become_designated_port(struct net_bridge_port *p);

/* br_stp_if.c */
extern void br_stp_enable_bridge(struct net_bridge *br);
extern void br_stp_disable_bridge(struct net_bridge *br);
extern void br_stp_enable_port(struct net_bridge_port *p);
extern void br_stp_disable_port(struct net_bridge_port *p);
extern void br_stp_recalculate_bridge_id(struct net_bridge *br);
extern void br_stp_set_bridge_priority(struct net_bridge *br,
				int newprio);
extern void br_stp_set_port_priority(struct net_bridge_port *p,
			      int newprio);
extern void br_stp_set_path_cost(struct net_bridge_port *p,
			  int path_cost);

/* br_stp_bpdu.c */
extern int br_stp_handle_bpdu(struct sk_buff *skb);

#endif
