/* vi: set sw=4 ts=4: */
/* dhcpd.c
 *
 * udhcp Server
 * Copyright (C) 1999 Matthew Ramsay <matthewr@moreton.com.au>
 *			Chris Trew <ctrew@moreton.com.au>
 *
 * Rewrite by Russ Dill <Russ.Dill@asu.edu> July 2001
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <fcntl.h>
#include <string.h>
#include <stdlib.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <netinet/in.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>
#include <signal.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <time.h>
#include <sys/time.h>

#include "debug.h"
#include "dhcpd.h"
#include "arpping.h"
#include "socket.h"
#include "options.h"
#include "files.h"
#include "leases.h"
#include "packet.h"
#include "serverpacket.h"
#include "pidfile.h"

#include <syslog.h>
#include "asyslog.h"

/* 32 bit change to 64 bit dennis 20080311 start */
#include "../../include/elbox_config.h"
#include <stdint.h>
/* 32 bit change to 64 bit dennis 20080311 end */

/* globals */
struct dhcpOfferedAddr *leases;
/* +++ Joy added static leases */
struct dhcpOfferedAddr *static_leases;
/* --- Joy added static leases */
struct server_config_t server_config;
static int signal_pipe[2];

/* Exit and cleanup */
static void exit_server(int retval)
{
	pidfile_delete(server_config.pidfile);
	CLOSE_LOG();
	exit(retval);
}

/* Signal handler */
static void signal_handler(int sig)
{
	if (send(signal_pipe[1], &sig, sizeof(sig), MSG_DONTWAIT) < 0)
	{
		LOG(LOG_ERR, "Could not send signal: %s", strerror(errno));
	}
}

int valid_ip(u_int32_t ipaddr,char *chaddr)
{
	u_int32_t addr;
	struct dhcpOfferedAddr *lease = NULL;
	int j;

	// 1.must in our pool
	addr = ntohl(ipaddr);
	if (addr < ntohl(server_config.start) || addr > ntohl(server_config.end))
	{
		DEBUG(LOG_INFO, "%s: out of range, addr=0x%08x, start=0x%08x, end=0x%08x",
			__FUNCTION__, addr, ntohl(server_config.start), ntohl(server_config.end));
		return 0;
	}

	// 2.is not server ip,broadcast,domain
	if((ipaddr & 0xFF)==0)
	{
		fprintf(stderr,"Error it's not a ip\n");
		return 0;
	}
	if((ipaddr & 0xFF) == 0xFF)
	{
		fprintf(stderr,"Error it's bcast \n");
		return 0;
	}
	if(ipaddr == server_config.server)
	{
		fprintf(stderr,"Error it's my ip \n");
		return 0;
	}

	// 3.check static table
	for (j = 0; (j < MAX_STATIC_LEASES) && (static_leases[j].yiaddr); j++)
	{
		if(ipaddr==static_leases[j].yiaddr)
		{
			if(memcmp(static_leases[j].chaddr, chaddr, sizeof(static_leases[j].chaddr))==0)
				return 2; // match static
			else
			{
				fprintf(stderr,"Error belong static \n");
				return 0;
			}
		}
	}

	// 4.check dynamic table
	if ((lease = find_lease_by_yiaddr(ipaddr)) && lease_expired(lease))
	{
		DEBUG(LOG_INFO, "%s: Woops !!!",__FUNCTION__);
		return 0;
	}

	// may other check
	return 1;
} 

#ifdef COMBINED_BINARY	
int udhcpd_main(int argc, char *argv[])
#else
int main(int argc, char *argv[])
#endif
{
	fd_set rfds;
	struct timeval tv;
	int server_socket = -1;
	int bytes, retval;
	struct dhcpMessage packet;
	unsigned char *state;
#if 0 /* Joy modified: hostname is moved to sendACK() */
	unsigned char *server_id, *requested, *hostname;
#else
	unsigned char *server_id, *requested;
#endif
	u_int32_t server_id_align, requested_align;
/* 32 bit change to 64 bit dennis 20080311 start */
#ifdef ELBOX_MODEL_DAP2590
	uint32_t timeout_end; 
#else
	unsigned long timeout_end;
#endif
/* 32 bit change to 64 bit dennis 20080311 end */
	struct option_set *option;
	struct dhcpOfferedAddr *lease;
	int pid_fd;
	int max_sock;
	int sig, ret=0;

	char mac[32];

	OPEN_LOG("udhcpd");
	LOG(LOG_INFO, "udhcp server (v%s) started", VERSION);

	/* +++ Joy added static leases */
	static_leases = malloc(sizeof(struct dhcpOfferedAddr) * MAX_STATIC_LEASES);
	memset(static_leases, 0, sizeof(struct dhcpOfferedAddr) * MAX_STATIC_LEASES);
	/* --- Joy added static leases */

	memset(&server_config, 0, sizeof(struct server_config_t));

	if (argc < 2) read_config(DHCPD_CONF_FILE);
	else read_config(argv[1]);

	pid_fd = pidfile_acquire(server_config.pidfile);
	pidfile_write_release(pid_fd);

	if ((option = find_option(server_config.options, DHCP_LEASE_TIME)))
	{
		memcpy(&server_config.lease, option->data + 2, 4);
		server_config.lease = ntohl(server_config.lease);
	}
	else
	{
		server_config.lease = LEASE_TIME;
	}

	leases = malloc(sizeof(struct dhcpOfferedAddr) * server_config.max_leases);
	memset(leases, 0, sizeof(struct dhcpOfferedAddr) * server_config.max_leases);
	read_leases(server_config.lease_file);
	/* +++ Joy added */
	write_leases();
	/* --- Joy added */

	if (read_interface(server_config.interface, &server_config.ifindex,
				&server_config.server, server_config.arp) < 0)
	{
		exit_server(1);
	}

#if 0
#ifndef DEBUGGING
	pid_fd = pidfile_acquire(server_config.pidfile); /* hold lock during fork. */
	if (daemon(0, 0) == -1)
	{
		perror("fork");
		exit_server(1);
	}
	pidfile_write_release(pid_fd);
#endif
#endif


	socketpair(AF_UNIX, SOCK_STREAM, 0, signal_pipe);
	signal(SIGUSR1, signal_handler);
	signal(SIGTERM, signal_handler);

	timeout_end = time(0) + server_config.auto_time;
	while(1)
	{	/* loop until universe collapses */
		if (server_socket < 0)
			if ((server_socket = listen_socket(INADDR_ANY, SERVER_PORT, server_config.interface)) < 0)
			{
				LOG(LOG_ERR, "FATAL: couldn't create server socket, %s", strerror(errno));
				exit_server(0);
			}			

		FD_ZERO(&rfds);
		FD_SET(server_socket, &rfds);
		FD_SET(signal_pipe[0], &rfds);
		if (server_config.auto_time)
		{
			tv.tv_sec = timeout_end - time(0);
			tv.tv_usec = 0;
		}
		if (!server_config.auto_time || tv.tv_sec > 0)
		{
			max_sock = server_socket > signal_pipe[0] ? server_socket : signal_pipe[0];
			retval = select(max_sock + 1, &rfds, NULL, NULL, server_config.auto_time ? &tv : NULL);
		}
		else
		{
			retval = 0; /* If we already timed out, fall through */
		}

		if (retval == 0)
		{
			write_leases();
			timeout_end = time(0) + server_config.auto_time;
			continue;
		}
		else if (retval < 0 && errno != EINTR)
		{
			DEBUG(LOG_INFO, "error on select");
			continue;
		}

		if (FD_ISSET(signal_pipe[0], &rfds))
		{
			if (read(signal_pipe[0], &sig, sizeof(sig)) < 0) continue; /* probably just EINTR */
			switch (sig)
			{
			case SIGUSR1:
				LOG(LOG_INFO, "Received a SIGUSR1");
				write_leases();
				/* why not just reset the timeout, eh */
				timeout_end = time(0) + server_config.auto_time;
				continue;
			case SIGTERM:
				LOG(LOG_INFO, "Received a SIGTERM");
				exit_server(0);
			}
		}

		if ((bytes = get_packet(&packet, server_socket)) < 0)
		{	/* this waits for a packet - idle */
			if (bytes == -1 && errno != EINTR)
			{
				DEBUG(LOG_INFO, "error on read, %s, reopening socket", strerror(errno));
				close(server_socket);
				server_socket = -1;
			}
			continue;
		}

		if ((state = get_option(&packet, DHCP_MESSAGE_TYPE)) == NULL)
		{
			DEBUG(LOG_ERR, "couldn't get option from packet, ignoring");
			continue;
		}

		/* ADDME: look for a static lease */
		lease = find_lease_by_chaddr(packet.chaddr);

		sprintf(mac, "%02x:%02x:%02x:%02x:%02x:%02x",
				packet.chaddr[0], packet.chaddr[1], packet.chaddr[2],
				packet.chaddr[3], packet.chaddr[4], packet.chaddr[5]);

		switch (state[0])
		{
		case DHCPDISCOVER:
			DEBUG(LOG_INFO,"received DISCOVER");
#ifndef LOGNUM
			syslog(ALOG_NOTICE|LOG_DEBUG, "DHCP: Server receive DISCOVER from %s.", mac);
#else
			syslog(ALOG_NOTICE|LOG_DEBUG, "NTC:018[%s]", mac);
#endif

			if (sendOffer(&packet) < 0)
			{
				LOG(LOG_ERR, "send OFFER failed");
			}
			break;

		case DHCPREQUEST:
			DEBUG(LOG_INFO, "%s: received REQUEST",__FUNCTION__);

			requested = get_option(&packet, DHCP_REQUESTED_IP);
			server_id = get_option(&packet, DHCP_SERVER_ID);

#ifndef LOGNUM
			syslog(ALOG_NOTICE|LOG_DEBUG, "DHCP: Server receive REQUEST from %s.", mac);
#else
			syslog(ALOG_NOTICE|LOG_DEBUG, "NTC:022[%s]", mac);
#endif

			if (requested)
			{
				memcpy(&requested_align, requested, 4);
				DEBUG(LOG_INFO, "%s: DHCP_REQUESTED_IP = 0x%08x",__FUNCTION__, requested_align);
			}
			if (server_id)
			{
				memcpy(&server_id_align, server_id, 4);
				DEBUG(LOG_INFO, "%s: DHCP_SERVER_ID = 0x%08x",__FUNCTION__, server_id_align);
			}

			/* +++ Joy added static lease */
			{
				struct dhcpOfferedAddr *static_lease;
				if ((static_lease = find_static_lease_by_chaddr(packet.chaddr)))
				{
					DEBUG(LOG_INFO, "%s: found static lease %02X:%02X:%02X:%02X:%02X:%02X",__FUNCTION__,
						packet.chaddr[0],packet.chaddr[1],packet.chaddr[2],packet.chaddr[3],packet.chaddr[4],packet.chaddr[5]);
					if (server_id && server_id_align == server_config.server)
					{
						if (requested && requested_align == static_lease->yiaddr)
						{
							/* Set the lease time to infinity */
							add_simple_option(packet.options, DHCP_LEASE_TIME, 0xFFFFFFFF);
							sendACK(&packet, static_lease->yiaddr);
						}
						else
						{
							sendNAK(&packet);
						}
						lease = NULL; //static_lease found, invalidate lease
					}
				}
			}
			/* --- Joy added static lease */
			if (lease)	/*ADDME: or static lease */
			{
				DEBUG(LOG_INFO, "%s: found lease !",__FUNCTION__);
				if (server_id)
				{
					/* SELECTING State */
					DEBUG(LOG_INFO, "%s: server_id = %08x",__FUNCTION__, ntohl(server_id_align));
					if (server_id_align == server_config.server &&
						requested && requested_align == lease->yiaddr)
						sendACK(&packet, lease->yiaddr);
				}
				else
				{
					if (requested)
					{
						/* INIT-REBOOT State */
						DEBUG(LOG_INFO, "%s: have requested!",__FUNCTION__);
						if (lease->yiaddr == requested_align)
							sendACK(&packet, lease->yiaddr);
						else
							sendNAK(&packet);
					}
					else
					{
						/* RENEWING or REBINDING State */
						if (lease->yiaddr == packet.ciaddr && valid_ip(packet.ciaddr, packet.chaddr))
							sendACK(&packet, lease->yiaddr);
						else
							sendNAK(&packet); /* don't know what to do!!!! */
					}						
				}
				/* what to do if we have no record of the client */
			}
			else if (server_id)
			{	/* SELECTING State */
				DEBUG(LOG_INFO, "%s: have server_id !",__FUNCTION__);
			}
			else if (requested)
			{
				DEBUG(LOG_INFO, "%s: have requested !",__FUNCTION__);
				if (find_static_lease_by_yiaddr(requested_align))
					sendNAK(&packet);
				else
				/* INIT-REBOOT State */
				if ((lease = find_lease_by_yiaddr(requested_align)))
				{
					if (lease_expired(lease))	/* probably best if we drop this lease */
						memset(lease->chaddr, 0, 16);
					else
						sendNAK(&packet);
				}
				else if(requested_align && (ret=valid_ip(requested_align, packet.chaddr)))
				{
					if(ret==2)	//-----static dhcp
						add_simple_option(packet.options, DHCP_LEASE_TIME, 0xFFFFFFFF);
					sendACK(&packet, requested_align);
				}
				else
					sendNAK(&packet);
			}
			else if(packet.ciaddr && (ret=valid_ip(packet.ciaddr, packet.chaddr)))
			{
				if(ret==2)	//-----static dhcp
					add_simple_option(packet.options, DHCP_LEASE_TIME, 0xFFFFFFFF);
				sendACK(&packet, packet.ciaddr);
			}
			else
				sendNAK(&packet);

			break;

		case DHCPDECLINE:
			DEBUG(LOG_INFO,"received DECLINE");
			if (lease)
			{
				memset(lease->chaddr, 0, 16);
				lease->expires = time(0) + server_config.decline_time;
			}			
			break;

		case DHCPRELEASE:
			DEBUG(LOG_INFO,"received RELEASE");
			if (lease)
			{
				lease->expires = time(0);
//+++joel 2004/7/2 04:04U when client release should re-write the lease file
				write_leases();
//---joel 2004/7/2 04:04U when client release should re-write the lease file
#ifndef LOGNUM
			syslog(ALOG_NOTICE|LOG_DEBUG, "DHCP: Server receive RELEASE from %s.", mac);
#else
			syslog(ALOG_NOTICE|LOG_DEBUG, "NTC:035[%s]", mac);
#endif
			}

			break;

		case DHCPINFORM:
			DEBUG(LOG_INFO,"received INFORM");
			send_inform(&packet);
			break;	

		default:
			LOG(LOG_WARNING, "unsupported DHCP message (%02x) -- ignoring", state[0]);
		}
	}

	return 0;
}
