/*
 * Copyright (C) 2000 Lennert Buytenhek
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <errno.h>
#include <asm/param.h>
#include "libbridge.h"
#include "brctl.h"

void br_cmd_addbr(struct bridge *br, char *brname, char *arg1)
{
	int err;

	if ((err = br_add_bridge(brname)) == 0)
		return;

	switch (err) {
	case EEXIST:
		fprintf(stderr,	"device %s already exists; can't create "
			"bridge with the same name\n", brname);
		break;

	default:
		perror("br_add_bridge");
		break;
	}
}

void br_cmd_delbr(struct bridge *br, char *brname, char *arg1)
{
	int err;

	if ((err = br_del_bridge(brname)) == 0)
		return;

	switch (err) {
	case ENXIO:
		fprintf(stderr, "bridge %s doesn't exist; can't delete it\n",
			brname);
		break;

	case EBUSY:
		fprintf(stderr, "bridge %s is still up; can't delete it\n",
			brname);
		break;

	default:
		perror("br_del_bridge");
		break;
	}
}

void br_cmd_addif(struct bridge *br, char *ifname, char *arg1)
{
	int err;
	int ifindex;

	ifindex = if_nametoindex(ifname);
	if (!ifindex) {
		fprintf(stderr, "interface %s does not exist!\n", ifname);
		return;
	}

	if ((err = br_add_interface(br, ifindex)) == 0)
		return;

	switch (err) {
	case EBUSY:
		fprintf(stderr,	"device %s is already a member of a bridge; "
			"can't enslave it to bridge %s.\n", ifname,
			br->ifname);
		break;

	case ELOOP:
		fprintf(stderr, "device %s is a bridge device itself; "
			"can't enslave a bridge device to a bridge device.\n",
			ifname);
		break;

	default:
		perror("br_add_interface");
		break;
	}
}

void br_cmd_delif(struct bridge *br, char *ifname, char *arg1)
{
	int err;
	int ifindex;

	ifindex = if_nametoindex(ifname);
	if (!ifindex) {
		fprintf(stderr, "interface %s does not exist!\n", ifname);
		return;
	}

	if ((err = br_del_interface(br, ifindex)) == 0)
		return;

	switch (err) {
	case EINVAL:
		fprintf(stderr, "device %s is not a slave of %s\n",
			ifname, br->ifname);
		break;

	default:
		perror("br_del_interface");
		break;
	}
}

void br_cmd_setageing(struct bridge *br, char *time, char *arg1)
{
	double secs;
	struct timeval tv;

	sscanf(time, "%lf", &secs);
	tv.tv_sec = secs;
	tv.tv_usec = 1000000 * (secs - tv.tv_sec);
	br_set_ageing_time(br, &tv);
}

void br_cmd_setbridgeprio(struct bridge *br, char *_prio, char *arg1)
{
	int prio;

	sscanf(_prio, "%i", &prio);
	br_set_bridge_priority(br, prio);
}

void br_cmd_setfd(struct bridge *br, char *time, char *arg1)
{
	double secs;
	struct timeval tv;

	sscanf(time, "%lf", &secs);
	tv.tv_sec = secs;
	tv.tv_usec = 1000000 * (secs - tv.tv_sec);
	br_set_bridge_forward_delay(br, &tv);
}

void br_cmd_setgcint(struct bridge *br, char *time, char *arg1)
{
	double secs;
	struct timeval tv;

	sscanf(time, "%lf", &secs);
	tv.tv_sec = secs;
	tv.tv_usec = 1000000 * (secs - tv.tv_sec);
	br_set_gc_interval(br, &tv);
}

void br_cmd_sethello(struct bridge *br, char *time, char *arg1)
{
	double secs;
	struct timeval tv;

	sscanf(time, "%lf", &secs);
	tv.tv_sec = secs;
	tv.tv_usec = 1000000 * (secs - tv.tv_sec);
	br_set_bridge_hello_time(br, &tv);
}

/* AP Client function 2007-01-21 start */
#ifdef APC_OPERATION_MODE
void br_cmd_apc(struct bridge *br, char *arg0, char *arg1)
{
	
	 int apc;
	 apc = 0;
	  if (!strcmp(arg0, "on") || !strcmp(arg0, "yes") || !strcmp(arg0, "1"))
		  apc = 1;
	  
	br_set_apc_state(br, apc);
}
#endif
/* AP Client function 2007-01-21 end */
/* IGMP Snooping dennis 2008-01-29 start */
#ifdef IGMP_SNOOP
void br_cmd_igmp(struct bridge *br, char *arg0, char *arg1)
{
	
	 int igmpenable;
	 igmpenable = 0;
	  if (!strcmp(arg0, "on") || !strcmp(arg0, "yes") || !strcmp(arg0, "1"))
		  igmpenable = 1;

	br_set_igmp_state(br, igmpenable);
}
#endif // IGMP_SNOOP
/* IGMP Snooping dennis 2008-01-29 end */
/* ethernet integration dennis2008-02-05 start */
#ifdef ETHLINK
void br_cmd_ethlink(struct bridge *br, char *arg0, char *arg1)
{
	
	 int ethlink;
	 ethlink = 0;
	  if (!strcmp(arg0, "on") || !strcmp(arg0, "yes") || !strcmp(arg0, "1"))
		  ethlink = 1;
	  
	br_set_ethlink_state(br, ethlink);
}
#endif
/* ethernet integration dennis2008-02-05 end */
/* e_partition 2008-02-1 start */
#ifdef E_PARTITION
void br_cmd_e_partition(struct bridge *br, char *arg0, char *arg1)
{
	int e_partition;
	e_partition = 0;
	if (!strcmp(arg0, "on") || !strcmp(arg0, "yes") || !strcmp(arg0, "1"))
		e_partition = 1;

	br_set_e_partition_state(br, e_partition);

}
#endif 
/* e_partition 2008-02-1 end */
/*dhcp server no provide to eth0's pc 2008-01-23 dennis start */
#ifdef DHCP_SERVER_ENABLE
void br_cmd_dhcp_server_enable(struct bridge *br, char *arg0, char *arg1)
{
	int dhcp_server_enable;
	dhcp_server_enable = 0;
	if (!strcmp(arg0, "on") || !strcmp(arg0, "yes") || !strcmp(arg0, "1"))
		dhcp_server_enable = 1;

	br_set_dhcp_server_enable(br, dhcp_server_enable);

}
#endif

/*+++ Limited Administration, Builder, 2008/03/11 +++*/
#ifdef LIMITED_ADMIN
/*limited administration type*/
void br_cmd_ladtype(struct bridge *br, char *arg0, char *arg1)
{
	int ladtype=0;
	if(!strcmp(arg0, "0")){
	    /*Limited Admin Disable*/
	    ladtype = 0;
	}else if(!strcmp(arg0, "1")){
	    /*Admin with VID*/
	    ladtype = 1;
	}else if(!strcmp(arg0, "2")){
	    /*Admin with Limited IP*/
	    ladtype = 2;
	}else if(!strcmp(arg0, "3")){
	    /*Both IP and VID*/
	    ladtype = 3;
	}else{
	    /*Default: Limited Admin Disabled*/
	    ladtype = 0;
	}

	br_set_ladtype(br, ladtype);

    return;
}

/*limited administration vlan ID*/
void br_cmd_ladvid(struct bridge *br, char *arg0, char *arg1)
{
    int vid=0;
    vid = atoi(arg0);
    
    br_set_ladvid(br, vid);
    
    return;
}

#include <netinet/in.h>
/*limited administration IP pool*/
void br_cmd_ladippool(struct bridge *br, char *arg1, char *arg2, char *arg3)
{
    unsigned int    poolidx=0;
    struct in_addr  ladipstart, ladipend;
#if 0
    printf("arg1: %s\n", arg1);
    printf("arg2: %s\n", arg2);
    printf("arg3: %s\n", arg3);
#endif /*0*/
    memset(&ladipstart, 0x0, sizeof(ladipstart));
    memset(&ladipend, 0x0, sizeof(ladipend));
    poolidx = atoi(arg1);
    if((poolidx<=4)&&(poolidx>=1))
        poolidx--;
    if(inet_aton(arg2, &ladipstart)&&inet_aton(arg3, &ladipend)){
#if 0
        printf("poolidx: %d\n", poolidx);
        printf("ladipstart.s_addr: %.8x\n", ladipstart.s_addr);
        printf("ladipend.s_addr: %.8x\n", ladipend.s_addr);
#endif /*0*/    
        br_set_ladippool(br, poolidx, ladipstart.s_addr, ladipend.s_addr);
    }
    return;
}
#endif /*LIMITED_ADMIN*/
/*+++ Limited Administration, Builder, 2008/03/11 +++*/

/* Jack add MULTI_VLAN +++ 11/04/07 */
#define VLAN_ID_MAX 4094
#define VLAN_ID_MIN 1
void br_cmd_primary_vlan_id(struct bridge *br, char *arg0, char *arg1)
{
	unsigned short primary_vlan_id;
	primary_vlan_id = 1; 
	primary_vlan_id = atoi(arg0);  
	if ( primary_vlan_id >= VLAN_ID_MIN-1 && primary_vlan_id <= VLAN_ID_MAX)
	{
	    printf("!!!br_cmd_primary_vlan_id, primary_vlan_id:%d \n", primary_vlan_id); 
	    br_set_primary_vlan_id(br, primary_vlan_id);
    }
    else
        printf("!!!br_cmd_primary_vlan_id, VID range should be %d-%d (%d means vlan disable)\n", VLAN_ID_MIN-1, VLAN_ID_MAX, VLAN_ID_MIN-1); 
}
#if 0
void br_cmd_show_primary_vlan_id(struct bridge *br, char *arg0, char *arg1)
{	
	br_dump_info_primary_vlan_id(br);	
}
#endif
/* Jack add MULTI_VLAN --- 11/04/07 */

/*dhcp server no provide to eth0's pc 2008-01-23 dennis end */
void br_cmd_setmaxage(struct bridge *br, char *time, char *arg1)
{
	double secs;
	struct timeval tv;

	sscanf(time, "%lf", &secs);
	tv.tv_sec = secs;
	tv.tv_usec = 1000000 * (secs - tv.tv_sec);
	br_set_bridge_max_age(br, &tv);
}

void br_cmd_setpathcost(struct bridge *br, char *arg0, char *arg1)
{
	int cost;
	struct port *p;

	if ((p = br_find_port(br, arg0)) == NULL) {
		fprintf(stderr, "can't find port %s in bridge %s\n", arg0, br->ifname);
		return;
	}

	sscanf(arg1, "%i", &cost);
	br_set_path_cost(p, cost);
}

void br_cmd_setportprio(struct bridge *br, char *arg0, char *arg1)
{
	int cost;
	struct port *p;

	if ((p = br_find_port(br, arg0)) == NULL) {
		fprintf(stderr, "can't find port %s in bridge %s\n", arg0, br->ifname);
		return;
	}

	sscanf(arg1, "%i", &cost);
	br_set_port_priority(p, cost);
}

void br_cmd_stp(struct bridge *br, char *arg0, char *arg1)
{
	int stp;

	stp = 0;
	if (!strcmp(arg0, "on") || !strcmp(arg0, "yes") || !strcmp(arg0, "1"))
		stp = 1;

	br_set_stp_state(br, stp);
}

void br_cmd_showstp(struct bridge *br, char *arg0, char *arg1)
{
	br_dump_info(br);
}

void br_cmd_show(struct bridge *br, char *arg0, char *arg1)
{
	printf("bridge name\tbridge id\t\tSTP enabled\tinterfaces\n");
	br = bridge_list;
	while (br != NULL) {
		printf("%s\t\t", br->ifname);
		br_dump_bridge_id((unsigned char *)&br->info.bridge_id);
		printf("\t%s\t\t", br->info.stp_enabled?"yes":"no");
		br_dump_interface_list(br);

		br = br->next;
	}
}

static int compare_fdbs(const void *_f0, const void *_f1)
{
	const struct fdb_entry *f0 = _f0;
	const struct fdb_entry *f1 = _f1;

#if 0
	if (f0->port_no < f1->port_no)
		return -1;

	if (f0->port_no > f1->port_no)
		return 1;
#endif

	return memcmp(f0->mac_addr, f1->mac_addr, 6);
}

void __dump_fdb_entry(struct fdb_entry *f)
{
	printf("%3i\t", f->port_no);
	printf("%.2x:%.2x:%.2x:%.2x:%.2x:%.2x\t",
	       f->mac_addr[0], f->mac_addr[1], f->mac_addr[2],
	       f->mac_addr[3], f->mac_addr[4], f->mac_addr[5]);
	printf("%s\t\t", f->is_local?"yes":"no");
	br_show_timer(&f->ageing_timer_value);
	printf("\n");
}

void br_cmd_showmacs(struct bridge *br, char *arg0, char *arg1)
{
	struct fdb_entry fdb[1024];
	int offset;

	printf("port no\tmac addr\t\tis local?\tageing timer\n");

	offset = 0;
	while (1) {
		int i;
		int num;

		num = br_read_fdb(br, fdb, offset, 1024);
		if (!num)
			break;

		qsort(fdb, num, sizeof(struct fdb_entry), compare_fdbs);

		for (i=0;i<num;i++)
			__dump_fdb_entry(fdb+i);

		offset += num;
	}
}

#ifdef MULTICAST_BWCTRL
void br_cmd_setbwctrl(struct bridge * br, char * arg0, char * arg1)
{
	int bandwidth;
	struct port * p;

	if ((p = br_find_port(br, arg0)) == NULL)
	{
		fprintf(stderr, "can't find port %s in bridge %s\n", arg0, br->ifname);
		return;
	}

	sscanf(arg1, "%i", &bandwidth);
	br_set_mlticst_bw(p, bandwidth);
}
#endif
#ifdef FORWARD_CTRL
void br_cmd_setfdctrl(struct bridge *br, char *arg0, char *arg1)
{
    int forward;
	struct port *p;

    if ((p = br_find_port(br, arg0)) == NULL)
    {
        fprintf(stderr, "can't find port %s in bridge %s\n", arg0, br->ifname);
        return;
    }

    forward = 1;
    if (!strcmp(arg1, "off") || !strcmp(arg1, "no") || !strcmp(arg1, "0"))
        forward = 0;

    br_set_forward_state(p, forward);
}
#endif
#ifdef REJECT_FORWARD_LIST
void br_cmd_addrejfwlist(struct bridge *br, char *arg0, char *arg1)
{
	struct port *p0, *p1;

	if ((p0 = br_find_port(br, arg0)) == NULL) {
		fprintf(stderr, "can't find port %s in bridge %s\n", arg0, br->ifname);
		return;
	}

	if ((p1 = br_find_port(br, arg1)) == NULL) {
		fprintf(stderr, "can't find port %s in bridge %s\n", arg1, br->ifname);
		return;
	}

	br_reject_forward_list(p0, p1, 1);	
	return;
}

void br_cmd_delrejfwlist(struct bridge *br, char *arg0, char *arg1)
{
	struct port *p0, *p1;

	if ((p0 = br_find_port(br, arg0)) == NULL) {
		fprintf(stderr, "can't find port %s in bridge %s\n", arg0, br->ifname);
		return;
	}

	if ((p1 = br_find_port(br, arg1)) == NULL) {
		fprintf(stderr, "can't find port %s in bridge %s\n", arg1, br->ifname);
		return;
	}

	br_reject_forward_list(p0, p1, 0);	
	return;
}
void br_cmd_shwrejfwlist(struct bridge *br, char *arg0, char *arg1)
{
	struct port *p0;

	if ((p0 = br_find_port(br, arg0)) == NULL) {
		fprintf(stderr, "can't find port %s in bridge %s\n", arg0, br->ifname);
		return;
	}

	br_reject_forward_list(p0, p0, 2);	
	return;
}
#endif

#ifdef AP_OPERATION_MODE
void br_cmd_apmode(struct bridge *br, char *arg0, char *arg1)
{
	int apmode;

	apmode = atol(arg0);
	br_set_apmode(br, apmode);
}
#endif /*AP_OPERATION_MODE*/

static struct command commands[] = {
	{0, 1, "addbr", br_cmd_addbr},
	{1, 1, "addif", br_cmd_addif},
	{0, 1, "delbr", br_cmd_delbr},
	{1, 1, "delif", br_cmd_delif},
	{1, 1, "setageing", br_cmd_setageing},
	{1, 1, "setbridgeprio", br_cmd_setbridgeprio},
	{1, 1, "setfd", br_cmd_setfd},
	{1, 1, "setgcint", br_cmd_setgcint},
	{1, 1, "sethello", br_cmd_sethello},
	{1, 1, "setmaxage", br_cmd_setmaxage},
	{1, 2, "setpathcost", br_cmd_setpathcost},
	{1, 2, "setportprio", br_cmd_setportprio},
	{0, 0, "show", br_cmd_show},
	{1, 0, "showmacs", br_cmd_showmacs},
	{1, 0, "showstp", br_cmd_showstp},
	{1, 1, "stp", br_cmd_stp},
#ifdef MULTICAST_BWCTRL
	{1, 2, "setbwctrl", br_cmd_setbwctrl},
#endif
#ifdef FORWARD_CTRL
	{1, 2, "setfdctrl", br_cmd_setfdctrl},
#endif
#ifdef REJECT_FORWARD_LIST
	{1,	2, "addrejfwlist", br_cmd_addrejfwlist},
	{1, 2, "delrejfwlist", br_cmd_delrejfwlist},
	{1, 1, "shwrejfwlist", br_cmd_shwrejfwlist},
#endif
#ifdef AP_OPERATION_MODE
	{1, 1, "apmode", br_cmd_apmode},
#endif /*AP_OPERATION_MODE*/
/* AP Client function 2007-01-21 start */
#ifdef APC_OPERATION_MODE
	{1, 1, "apc", br_cmd_apc},
#endif	
/* AP Client function 2007-01-21 end */
/*dhcp server no provide to eth0's pc 2008-01-23 dennis start */
#ifdef DHCP_SERVER_ENABLE
	{1, 1, "dhcp_server_enable", br_cmd_dhcp_server_enable},
#endif
/*dhcp server no provide to eth0's pc 2008-01-23 dennis end */
/* IGMP Snooping dennis 2008-01-29 start */
#ifdef IGMP_SNOOP
	{1, 1, "igmp_snooping", br_cmd_igmp}, 
#endif //IGMP_SNOOP	
/* IGMP Snooping dennis 2008-01-29 end */
/* e_partition 2008-02-1 start */
#ifdef E_PARTITION
	{1, 1, "e_partition", br_cmd_e_partition},
#endif	
/* e_partition 2008-02-1 end */

	//{1, 0, "showprimaryvid", br_cmd_show_primary_vlan_id},
	{1, 1, "setprimaryvid", br_cmd_primary_vlan_id},  /* Jack add MULTI_VLAN 11/04/07 */
/* ethernet integration dennis2008-02-05 start */
#ifdef ETHLINK
	{1, 1, "ethlink", br_cmd_ethlink},
#endif	
/* ethernet integration dennis2008-02-05 end */
/*+++ Limited Administration, Builder, 2008/03/11 +++*/
#ifdef LIMITED_ADMIN
/*limited administration type*/
    {1, 1, "ladtype",       br_cmd_ladtype},
/*limited administration vlan ID*/
    {1, 1, "ladvid",        br_cmd_ladvid},
/*limited administration IP start from*/
    {1, 3, "ladippool",    br_cmd_ladippool},
#endif /*LIMITED_ADMIN*/
/*+++ Limited Administration, Builder, 2008/03/11 +++*/
};

struct command *br_command_lookup(char *cmd)
{
	int i;
	int numcommands;

	numcommands = sizeof(commands)/sizeof(commands[0]);

	for (i=0;i<numcommands;i++)
		if (!strcmp(cmd, commands[i].name))
			return &commands[i];

	return NULL;
}
